#include "Camera.h"
#include <GL/glut.h>

using namespace MyAlgebra;

Camera::Camera()
: m_EyeDistance(5.f), m_TrackballRadius(1.f), m_WindowWidth(256), m_WindowHeight(256),
m_FieldOfView(30.f), m_NearZ(0.1f), m_FarZ(100.f)
{
	reset();
}

Camera::Camera(float _eyeDistance, int _width, int _height)
: m_EyeDistance(_eyeDistance), m_WindowWidth(_width), m_WindowHeight(_height),
m_FieldOfView(30.f), m_NearZ(0.1f), m_FarZ(100.f)
{
	// This size should really be based on the distance from the center of
	// rotation to the point on the object underneath the mouse.  That
	// point would then track the mouse as closely as possible.  This is a
	// simple example, though, so that is left as an exercise.
	m_TrackballRadius = 1.0f;
	reset();
}

// Bring trackcamera to home position, no translation, zero rotation.
void Camera::reset()
{
	m_QuatRot = quat(0.0, 0.0, 0.0, 1.0);
	m_QuatIncrement = quat(0.0, 0.0, 0.0, 1.0);
	m_Translation = vec3(0.0, 0.0, -m_EyeDistance);
}

void Camera::grab(CameraAction act, int xp, int yp)
{
	float x = scalePoint(xp, 0, m_WindowWidth);
	float y = scalePoint(yp, m_WindowHeight, -(long)m_WindowHeight);

	stopSpin();
	m_CurrentAction = act;

	m_LastX = x;
	m_LastY = y;
}

void Camera::roll(int xp, int yp)
{
	float x = scalePoint(xp, 0, m_WindowWidth);
	float y = scalePoint(yp, m_WindowHeight, -(long)m_WindowHeight);

	switch (m_CurrentAction)
	{
	case ROTATE_ACTION:
		rotate(x, y);
		break;
	case PAN_ACTION:
		pan(x, y);
		break;
	case ZOOM_ACTION:
		zoom(x, y);
		break;
	}

	m_LastX = x;	/* remember for next time */
	m_LastY = y;
}

mat4 Camera::getModelviewMatrix() const
{
	mat4 M = mat4_id;

#if 0
	m_QuatRot.to4x4Matrix( M.mat_array );
#else
	mat3 R;
	m_QuatRot.to3x3Matrix( R.mat_array );
	M.a00 = R.a00;	M.a01 = R.a01;	M.a02 = R.a02;	M.a03 = 0.f;
	M.a10 = R.a10;	M.a11 = R.a11;	M.a12 = R.a12;	M.a13 = 0.f;
	M.a20 = R.a20;	M.a21 = R.a21;	M.a22 = R.a22;	M.a23 = 0.f;
	M.a30 = 0.f;	M.a31 = 0.f;	M.a32 = 0.f;	M.a33 = 1.f;
#endif

	M.a03 = m_Translation.x;
	M.a13 = m_Translation.y;
	M.a23 = m_Translation.z;

	return M;
}

void Camera::getModelviewMatrix(float M[16]) const
{
#if 0
	m_QuatRot.to4x4Matrix(M);
#else
	mat3 R;
	m_QuatRot.to3x3Matrix( R.mat_array );
	M[0] = R.a00;	M[4] = R.a01;	M[8] = R.a02;	M[12] = 0.f;
	M[1] = R.a10;	M[5] = R.a11;	M[9] = R.a12;	M[13] = 0.f;
	M[2] = R.a20;	M[6] = R.a21;	M[10] = R.a22;	M[14] = 0.f;
	M[3] = 0.f;		M[7] = 0.f;		M[11] = 0.f;	M[15] = 1.f;
#endif

	M[12] = m_Translation.x;
	M[13] = m_Translation.y;
	M[14] = m_Translation.z;
}

void Camera::setModelviewMatrix(float M[16])
{
	m_QuatRot.from4x4Matrix(M);

	m_Translation.x = M[12];
	m_Translation.y = M[13];
	m_Translation.z = M[14];
}

// sets up a orthonormal system, compatible with gluLookAt()
void Camera::setLookAt(const vec3 &eye, const vec3 &center, const vec3 &up)
{
	vec3 forward = center - eye;
	forward.normalize();

	vec3 side = forward ^ up;
	normalize(side);

	const vec3 newUp = side ^ forward;

	float R[9];

	R[0] = side.x;
	R[1] = side.y;
	R[2] = side.z;

	R[3] = newUp.x;
	R[4] = newUp.y;
	R[5] = newUp.z;

	R[6] = -forward.x;
	R[7] = -forward.y;
	R[8] = -forward.z;

	//std::cerr << __FUNCTION__ << ": R before quat conversion: " << mat3(R) << std::endl;

	m_QuatRot.from3x3Matrix(R);

	//std::cerr << __FUNCTION__ << ": m_QuatRot = " << m_QuatRot << std::endl;

	//float M[9];
	//m_QuatRot.to3x3Matrix(M);
	//std::cerr << __FUNCTION__ << ": R after quat conversion: " << mat3(M) << std::endl;

#if 1
	m_Translation.x = -(R[0]*eye.x + R[3]*eye.y + R[6]*eye.z);
	m_Translation.y = -(R[1]*eye.x + R[4]*eye.y + R[7]*eye.z);
	m_Translation.z = -(R[2]*eye.x + R[5]*eye.y + R[8]*eye.z);
#else
	m_Translation.x = -dot(side, eye);
	m_Translation.y = -dot(newUp, eye);
	m_Translation.z =  dot(forward, eye);	// this should be negated?
#endif
}

void Camera::getLookAt(vec3 &eye, vec3 &center, vec3 &up) const
{
	float R[9];
	m_QuatRot.to3x3Matrix(R);

#if 1
	eye.x = -(R[0]*m_Translation.x+R[1]*m_Translation.y+R[2]*m_Translation.z);
	eye.y = -(R[3]*m_Translation.x+R[4]*m_Translation.y+R[5]*m_Translation.z);
	eye.z = -(R[6]*m_Translation.x+R[7]*m_Translation.y+R[8]*m_Translation.z);

	const vec3 forward(-R[6], -R[7], -R[8]);

	center = forward + eye;

	up.set( R[3], R[4], R[5] );
#else
	eye.x = -(R[0]*m_Translation.x+R[3]*m_Translation.y+R[6]*m_Translation.z);
	eye.y = -(R[1]*m_Translation.x+R[4]*m_Translation.y+R[7]*m_Translation.z);
	eye.z = -(R[2]*m_Translation.x+R[5]*m_Translation.y+R[8]*m_Translation.z);

	const vec3 forward(-R[6], -R[7], -R[8]);
	const vec3 side(R[0], R[1], R[2]);

	center = forward + eye;
	up = side ^ forward;
#endif
}

// returns camera's eye position
vec3 Camera::getEyePosition() const
{
	vec3 eye;

	// translation multiplied by the inverse of the rotation matrix
	// note: inverse of rotation == transpose of rotation

	float R[9];
	m_QuatRot.to3x3Matrix(R);

#if 0
	eye.x = -(R[0]*m_Translation.x+R[1]*m_Translation.y+R[2]*m_Translation.z);
	eye.y = -(R[3]*m_Translation.x+R[4]*m_Translation.y+R[5]*m_Translation.z);
	eye.z = -(R[6]*m_Translation.x+R[7]*m_Translation.y+R[8]*m_Translation.z);
#else
	eye.x = -(R[0]*m_Translation.x+R[3]*m_Translation.y+R[6]*m_Translation.z);
	eye.y = -(R[1]*m_Translation.x+R[4]*m_Translation.y+R[7]*m_Translation.z);
	eye.z = -(R[2]*m_Translation.x+R[5]*m_Translation.y+R[8]*m_Translation.z);
#endif

	return eye;
}

// returns orthonormal base, in a left-handed system
void Camera::getModelviewCoordinates(vec3 &xAxis, vec3 &yAxis, vec3 &zAxis) const
{
	float R[9];
	m_QuatRot.to3x3Matrix(R);

#if 1
	xAxis.set( R[0], R[1], R[2] );
	yAxis.set( R[3], R[4], R[5] );
	zAxis.set( R[6], R[7], R[8] );
#else
	xAxis.set( R[0], R[3], R[6] );
	yAxis.set( R[1], R[4], R[7] );
	zAxis.set( R[2], R[5], R[8] );
#endif
}

void Camera::setGLModelview() const
{
	float M[16];

#if 1

	glTranslatef(m_Translation.x, m_Translation.y, m_Translation.z);

	m_QuatRot.getMatrix(M);	// ɖH

#else

#if 0
	glTranslatef(m_Translation.x, m_Translation.y, m_Translation.z);

	mat3 R;
	m_QuatRot.to3x3Matrix( R.mat_array );
	M[0] = R.a00;	M[4] = R.a01;	M[8 ] = R.a02;	M[12] = 0.f;
	M[1] = R.a10;	M[5] = R.a11;	M[9 ] = R.a12;	M[13] = 0.f;
	M[2] = R.a20;	M[6] = R.a21;	M[10] = R.a22;	M[14] = 0.f;
	M[3] = 0.f;		M[7] = 0.f;		M[11] = 0.f;	M[15] = 1.f;
#endif
#if 0
	mat3 R;
	m_QuatRot.to3x3Matrix( R.mat_array );
	M[0] = R.a00;	M[4] = R.a01;	M[8 ] = R.a02;	M[12] = m_Translation.x;
	M[1] = R.a10;	M[5] = R.a11;	M[9 ] = R.a12;	M[13] = m_Translation.y;
	M[2] = R.a20;	M[6] = R.a21;	M[10] = R.a22;	M[14] = m_Translation.z;
	M[3] = 0.f;		M[7] = 0.f;		M[11] = 0.f;	M[15] = 1.f;
#endif

#endif

	glMultMatrixf(M);
}

void Camera::setGLPerspective() const
{
	gluPerspective(m_FieldOfView, m_WindowWidth/(float)m_WindowHeight, m_NearZ, m_FarZ);
}


// private functions ---------------------------------------------------------------------

void Camera::calcRotation(float newX, float newY, float oldX, float oldY) 
{
	vec3 p1, p2;	/* 3D mouse points	*/
	float	L;		/* sin^2(2 * phi)	*/
	vec3 q;

	// Check for zero rotation
	if (newX == oldX && newY == oldY) {
		m_QuatIncrement.set(0.0, 0.0, 0.0, 1.0);
		return;
	}

	// Form two vectors based on input points, find rotation axis
	p1 = vec3(newX, newY, projectToSphere(newX, newY));
	p2 = vec3(oldX, oldY, projectToSphere(oldX, oldY));

	q.cross(p1, p2);	/* axis of rotation from p1 and p2 */

	L = q.sq_norm() / (p1.sq_norm() * p2.sq_norm());
	L = sqrt(1 - L);

	q.normalize();		/* q' = axis of rotation */
	q.scale(sqrt((1 - L)/2));	/* q' = q' * sin(phi) */

	m_QuatIncrement.set(-q.x, -q.y, -q.z, sqrt((1 + L)/2));
	//m_QuatIncrement.set(q.x, q.y, q.z, sqrt((1 + L)/2));
	//	qinc.set(q.x, q.y, q.z, sqrt((1 + L)/2));
}

// Project an x,y pair onto a sphere of radius r OR a hyperbolic sheet
// if we are away from the center of the sphere.
float Camera::projectToSphere(float x, float y) 
{
	float d = sqrtf(x*x + y*y);

	// Inside sphere 
	if (d < m_TrackballRadius * 0.70710678118654752440)
	{
		return sqrt(m_TrackballRadius*m_TrackballRadius - d*d);
	}
	// On hyperbola 
	else
	{
		return (m_TrackballRadius * m_TrackballRadius) / (2.0f * d);
	}
}

// Scales integer point to the range [-1, 1]
float Camera::scalePoint(long pt, long origin, long size)
{
	float x = (float) (pt - origin) / (float) size;

	if (x < 0.0f) x = 0.0f;
	if (x > 1.0f) x = 1.0f;

	return (2.0f * x) - 1.0f;
}

//void Camera::testDump()
//{
//	mat3 R;
//	m_QuatRot.to3x3Matrix( R.mat_array );
//	std::cerr << __FUNCTION__ << ": 3x3 mat: " << R << std::endl;
//
//
//	mat4 M;
//	m_QuatRot.to4x4Matrix( M.mat_array );
//	std::cerr << __FUNCTION__ << ": 4x4 mat: " << M << std::endl;
//}
