#ifndef _CAMERA_H_
#define _CAMERA_H_

#include "my_algebra.h"

class Camera {
public:
	enum CameraAction
	{
		NO_ACTION = 0,
		ROTATE_ACTION,
		PAN_ACTION,
		ZOOM_ACTION
	};

	Camera();
	Camera(float _eyeDistance, int _width, int _height);

	/// Call to reset the Camera and bring everything home to square one.
	void reset();

	// Modelview -----------------------------------------------------------------

	MyAlgebra::mat4 getModelviewMatrix() const;
	void getModelviewMatrix(float M[16]) const;
	void setModelviewMatrix(float M[16]);

	// sets up a orthonormal system, compatible with gluLookAt()
	void setLookAt(const MyAlgebra::vec3 &eye, const MyAlgebra::vec3 &center, const MyAlgebra::vec3 &up);
	void getLookAt(MyAlgebra::vec3 &eye, MyAlgebra::vec3 &center, MyAlgebra::vec3 &up) const;

	// returns camera's eye position
	MyAlgebra::vec3 getEyePosition() const;

	// returns orthonormal base, in a left-handed system
	void getModelviewCoordinates(MyAlgebra::vec3 &xAxis, MyAlgebra::vec3 &yAxis, MyAlgebra::vec3 &zAxis) const;

	void setGLModelview() const;

	// OpenGL-related --------------------------------------------------------------

	MyAlgebra::Real calcScreenDistance() const
	{
		return 0.5f * m_WindowHeight / tanf(m_FieldOfView * (float)M_PI / 360.f);
	}

	int getWindowWidth() const { return m_WindowWidth; }
	int getWindowHeight() const { return m_WindowHeight; }

	void setFieldOfView(MyAlgebra::Real f) { m_FieldOfView = f; }
	inline MyAlgebra::Real getFieldOfView() const { return m_FieldOfView; }

	void setNearZ(MyAlgebra::Real z) { m_NearZ = z; }
	inline MyAlgebra::Real getNearZ() const { return m_NearZ; }

	void setFarZ(MyAlgebra::Real z) { m_FarZ = z; }
	inline MyAlgebra::Real getFarZ() const { return m_FarZ; }

	void setGLPerspective() const;

	// UI -----------------------------------------------------------

	/// Call this function to start action when mouse button is pressed 
	void grab(CameraAction,int,int);

	void roll(int xp, int yp);

	/// Call this function to stop action when mouse is released.
	inline void release() { m_CurrentAction = NO_ACTION; }

	/// Returns true if the camera is `grabbed' and not released yet.
	bool isGrabbed() const { return (m_CurrentAction != NO_ACTION); }

	// Misc --------------------------------------------------------------

	/** Spin. used both to spin while button is pressed and if the camera is just
		spinning while program is idling. */
	// Rotationaly spin the trackcamera by the current increment.
	// Use this to implement rotational glide.
	inline void doSpin() { m_QuatRot.mul(m_QuatIncrement); }
	
	// Zeroes the rotation value - makes everything stop.
	// Cease any rotational glide by zeroing the increment.
	inline void stopSpin() { 	m_QuatIncrement.set(0.0, 0.0, 0.0, 1.0); }

	void setWindowSize(unsigned _width, unsigned _height)
	{
		m_WindowWidth = _width;
		m_WindowHeight = _height;
	}

	//void testDump();

private:
	int m_WindowWidth, m_WindowHeight;

	MyAlgebra::quat	m_QuatRot;
	MyAlgebra::quat	m_QuatIncrement;

	MyAlgebra::vec3 m_Translation;

	float m_TrackballRadius;
	float m_EyeDistance;
	float m_LastX, m_LastY;

	// to be used by OpenGL
	MyAlgebra::Real m_FieldOfView;
	MyAlgebra::Real m_NearZ, m_FarZ;

	CameraAction m_CurrentAction;

	// private functions ------------------------------------------------

	inline void rotate(float newx, float newy)
	{
		calcRotation(m_LastX, m_LastY, newx, newy);
		doSpin();
	}

	inline void pan(float newx, float newy)
	{
		m_Translation.x += (newx - m_LastX) * m_EyeDistance;
		m_Translation.y += (newy - m_LastY) * m_EyeDistance;
	}

	inline void zoom(float newx, float newy)
	{
		m_Translation.z += (newy - m_LastY) * m_EyeDistance;
	}

	void calcRotation(float, float, float, float);

	float projectToSphere(float, float);

	float scalePoint(long pt, long origin, long size);

};

#endif // __CAMERA_H__
