#include "CameraViewWindow.h"
#include "Camera.h"
#include "Scene.h"
#include "ImageRect.h"
#include "RecursiveRayTracer.h"
#include "EnvironmentMap.h"

using namespace std;
using namespace MyAlgebra;

extern RayTracer *g_pRayTracer;
extern ImageRect<vec3> g_ScreenBuffer;
extern Camera g_Camera;
extern Scene g_Scene;

extern bool g_ScreenUpdateRequired;
extern bool g_DisplayRayTracingResult;
extern int g_KeepRayTracing;

extern bool g_SingleRayDebug;
extern vector<int> g_TracedGridCells;

void CameraViewWindow::display()
{
	if ( ! m_pImageBuffer) return;

	if (g_KeepRayTracing && g_ScreenUpdateRequired)
	{
		g_pRayTracer->renderScene( g_ScreenBuffer );

		g_DisplayRayTracingResult = true;
		g_ScreenUpdateRequired = false;
	}

	glPushAttrib(GL_ENABLE_BIT);

	glClearColor(1,1,1,1);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	const int w = m_pImageBuffer->getWidth();
	const int h = m_pImageBuffer->getHeight();

	if (g_DisplayRayTracingResult)
	{
		glDrawPixels(w, h, GL_RGB, GL_FLOAT, m_pImageBuffer->getData());
	}
	else
	{
		glEnable(GL_DEPTH_TEST);

		glMatrixMode(GL_PROJECTION);
		glPushMatrix();
		glLoadIdentity();
		g_Camera.setGLPerspective();

		glMatrixMode(GL_MODELVIEW);
		glPushMatrix();
		glLoadIdentity();
		g_Camera.setGLModelview();
		//glMultMatrixf( g_Camera.getModelviewMatrix() );

		if ( g_Scene.isEnvironmentMapAvailable() )
		{
			EnvironmentMap *envmap = g_Scene.getEnvironmentMap();

			if ( ! envmap->isTextureReady())
				envmap->makeGLTextures();
			envmap->drawSkyBox();
		}

		for (int i=0; i<g_Scene.getNumObjects(); i++)
		{
			glColor3fv( g_Scene.getPseudoColor(i) );
			g_Scene.getObject(i)->drawGL();
		}

		glMatrixMode(GL_PROJECTION);
		glPopMatrix();

		glMatrixMode(GL_MODELVIEW);
		glPopMatrix();

		// preview message

		glColor3f(1,1,0);
		displayMessage(10, 20, "OpenGL Preview");
	}

	glPopAttrib();

	glutSwapBuffers();
}

void CameraViewWindow::displayMessage(int xi, int yi, const char *message)
{
	glPushAttrib(GL_ENABLE_BIT);
	glDisable(GL_BLEND);
	glDisable(GL_DEPTH_TEST);
	glDisable(GL_LIGHTING);

	const int w = g_Camera.getWindowWidth();
	const int h = g_Camera.getWindowHeight();

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();

	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	gluOrtho2D(0, w, 0, h);

	glRasterPos2i(xi, yi);

	for (unsigned int i=0; message[i] != '\0'; i++)
		glutBitmapCharacter(GLUT_BITMAP_HELVETICA_18, message[i]);

	glRasterPos2i(0, 0);

	glPopMatrix();

	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();

	glPopAttrib();
}

void CameraViewWindow::mouse( int button, int state, int x, int _y )
{
	if ( state == GLUT_DOWN && (GLUT_ACTIVE_SHIFT&glutGetModifiers()) )
	{
		g_SingleRayDebug = true;
		g_TracedGridCells.clear();

		const int y = m_WindowHeight - _y;

		cerr << __FUNCTION__ << ": debugging ray at (" << x << "," << y << ")" << endl;

		((RecursiveRayTracer *)g_pRayTracer)->singleRayTrace( g_ScreenBuffer, x, y );

		g_SingleRayDebug = false;

		return;
	}

	if ( ! g_Camera.isGrabbed() && state == GLUT_DOWN)
	{
		if (button == GLUT_LEFT_BUTTON)
			g_Camera.grab(Camera::ROTATE_ACTION, x, _y);
		else if (button == GLUT_MIDDLE_BUTTON)
			g_Camera.grab(Camera::ZOOM_ACTION, x, _y);
		else if (button == GLUT_RIGHT_BUTTON)
			g_Camera.grab(Camera::PAN_ACTION, x, _y);
	}
	else if (state == GLUT_UP)
	{
		if (g_Camera.isGrabbed()) g_Camera.release();
		display();
	}
}

void CameraViewWindow::motion( int x, int _y )
{
	if (g_Camera.isGrabbed())
	{
		g_Camera.roll(x,_y);
		Display();

		g_DisplayRayTracingResult = false;
		g_ScreenUpdateRequired = true;
	}
}

void CameraViewWindow::reshape( int w, int h )
{
	if ( h < 1 ) return;

	glViewport(0, 0, w, h);

	const double cameraFovy = g_Camera.getFieldOfView();
	const double zNear = g_Camera.getNearZ();
	const double zFar = g_Camera.getFarZ();
	const double aspectRatio = w/(double)h;

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(cameraFovy, aspectRatio, zNear, zFar);

	g_ScreenBuffer.allocate(w, h);

	g_ScreenUpdateRequired = true;

	g_Camera.setWindowSize(w,h);
}

void CameraViewWindow::keyboard( unsigned char key, int x, int y )
{
	switch (key)
	{
	case 'C':
		{
			vec3 eye, center, up;
			g_Camera.getLookAt(eye, center, up);

			cout << "camera: " << endl
				<< eye << endl << center << endl << up << endl;
		}
		break;
	default:
		break;
	}
}

