//#define _CRT_SECURE_NO_WARNINGS

#include "EnvironmentMap.h"
#include "data_path.h"
#include "rgbe.h"

using namespace std;
using namespace MyAlgebra;

vec3 EnvironmentMap::fetchColor(const vec3 &dir) const
{
	if ( ! m_pData) return vec3(0,0,0);

	const int facePixels = m_FaceWidth *m_FaceHeight;

	if ((fabsf(dir.x) >= fabsf(dir.y)) && (fabsf(dir.x) >= fabsf(dir.z)))
	{
		if (dir.x < 0.0f)	// right face (positive x)
		{
			vec3 *addr = &m_pData[0];
			return readTexture(addr, (dir.z / dir.x+ 1.0f) * 0.5f, 1.0f - (dir.y / dir.x+ 1.0f) * 0.5f);
		}
		else // left face (negative x)
		{
			vec3 *addr = &m_pData[facePixels];
			return readTexture(addr, (dir.z / dir.x+ 1.0f) * 0.5f, ( dir.y / dir.x + 1.0f) * 0.5f);
		}
	}
	else if ((fabsf(dir.y) >= fabsf(dir.x)) && (fabsf(dir.y) >= fabsf(dir.z)))
	{
		// TODO: fill here to sample the top (positive-y) face and the bottom face (negative-y)

		return vec3(0.f);
	}
	else if ((fabsf(dir.z) >= fabsf(dir.x)) && (fabsf(dir.z) >= fabsf(dir.y)))
	{
		// TODO: fill here to sample the forward (negative-z) face and the backward face (positive-z)

		return vec3(0.f);
	}

	return vec3(0.f);
}

vec3 EnvironmentMap::readTexture(const vec3 *addr, float u, float v) const
{
	// TODO: rewrite this code using bilinear interpolation

	const int x0 = max(0, min((int)(u * m_FaceWidth), m_FaceWidth-1));
	const int y0 = max(0, min((int)(v * m_FaceHeight), m_FaceHeight-1));

	return addr[x0 + m_FaceWidth*y0];
}

bool EnvironmentMap::load(const char *filename)
{
	const string extension = data_path::getExtension(filename);

	data_path imagePath;
	imagePath.path.push_back(".");
	imagePath.path.push_back("Cubemap");
	imagePath.path.push_back("../Cubemap");
	imagePath.path.push_back("..");
	imagePath.path.push_back("../..");

	if (extension == "hdr")
	{
		return loadFromCrossHDR( imagePath.get_file(filename).c_str() );
	}
	else
	{
		cerr << __FUNCTION__ << ": format not supported" << endl;
		return false;
	}
}

void EnvironmentMap::convertToLongitudeLatitudeFormat(int height, ImageRGBf &out) const
{
	const int width = 2 * height;

	const float hh = 0.5f * height;
	const float hw = 0.5f * width;

	const float invGamma = 1.f/m_HdrGamma;

	out.allocate(width,height);

	for (int yi=0; yi<height; yi++)
	{
		vec3 *addr = out.getData() + yi*width;
		
		const float lat = M_PI * (height - yi - 0.5f) / (float)height;

		for (int xi=0; xi<width; xi++)
		{
			const float lon = M_PI * (xi + 0.5f - hw) / hw;

			const vec3 dir( sinf(lat)*sinf(lon), cosf(lat), -sinf(lat)*cosf(lon) );

			vec3 color = fetchColor(dir);

			color.r = powf(color.r, invGamma);
			color.g = powf(color.g, invGamma);
			color.b = powf(color.b, invGamma);

			addr[xi] = color;
		}
	}
}

bool EnvironmentMap::loadFromCrossHDR(const char *filename)
{
	FILE *fp = fopen(filename, "rb");
	if ( ! fp)
	{
		cerr << __FUNCTION__ << ": cannot open file: " << filename << endl;
		return false;
	}

	rgbe_header_info header;

	if (RGBE_ReadHeader(fp, &m_ImageWidth, &m_ImageHeight, &header))
	{
		cerr << __FUNCTION__ << ": cannot read HDR header: " << filename << endl;
		return false;
	}

	unsigned char *rawData = new unsigned char[m_ImageWidth * m_ImageHeight * 4];

	if ( ! rawData)
	{
		cerr << __FUNCTION__ << ": cannot malloc data" << endl;
		return false;
	}

	if (RGBE_ReadPixels_Raw_RLE(fp, rawData, m_ImageWidth, m_ImageHeight))
	{
		cerr << __FUNCTION__ << ": cannot copy HDR data: " << filename << endl;
		delete [] rawData;
		return false;
	}

	// begin reading HDR data

	// cross is 3 faces wide, 4 faces high
	m_FaceWidth = m_ImageWidth / 3;
	m_FaceHeight = m_ImageHeight / 4;

	const int facePixels = m_FaceWidth * m_FaceHeight;

	if (m_pData) delete [] m_pData;
	m_pData = new vec3[6 * facePixels];
	memset(m_pData, 0, 6 * facePixels * sizeof(vec3));

	// extract 6 faces

	// left face (negative x)

	for (int yi=0; yi<m_FaceHeight; yi++)
	{
		for (int xi=0; xi<m_FaceWidth; xi++)
		{
			unsigned char *src = getPixelAddr(rawData, xi, 2*m_FaceHeight+yi);
			float r, g, b;
			rgbe2float(&r, &g, &b, src);
			r = powf(r, m_HdrGamma);
			g = powf(g, m_HdrGamma);
			b = powf(b, m_HdrGamma);
			m_pData[xi + m_FaceWidth*yi].set( r, g, b );
		}
	}

	// right face (positive x)

	for (int yi=0; yi<m_FaceHeight; yi++)
	{
		for (int xi=0; xi<m_FaceWidth; xi++)
		{
			unsigned char *src = getPixelAddr(rawData, 2*m_FaceWidth+xi, 2*m_FaceHeight+yi);
			float r, g, b;
			rgbe2float(&r, &g, &b, src);
			r = powf(r, m_HdrGamma);
			g = powf(g, m_HdrGamma);
			b = powf(b, m_HdrGamma);
			m_pData[facePixels + xi + m_FaceWidth*yi].set( r, g, b );
		}
	}

	// top face (positive y)

	for (int yi=0; yi<m_FaceHeight; yi++)
	{
		for (int xi=0; xi<m_FaceWidth; xi++)
		{
			unsigned char *src = getPixelAddr(rawData, m_FaceWidth+xi, 3*m_FaceHeight+yi);
			float r, g, b;
			rgbe2float(&r, &g, &b, src);
			r = powf(r, m_HdrGamma);
			g = powf(g, m_HdrGamma);
			b = powf(b, m_HdrGamma);
			m_pData[2*facePixels + xi + m_FaceWidth*yi].set( r, g, b );
		}
	}

	// bottom face (negative y)

	for (int yi=0; yi<m_FaceHeight; yi++)
	{
		for (int xi=0; xi<m_FaceWidth; xi++)
		{
			//unsigned char *src = getPixelAddr(rawData, m_FaceWidth+xi, 2*m_FaceHeight-1-yi);
			unsigned char *src = getPixelAddr(rawData, m_FaceWidth+xi, m_FaceHeight+yi);
			float r, g, b;
			rgbe2float(&r, &g, &b, src);
			r = powf(r, m_HdrGamma);
			g = powf(g, m_HdrGamma);
			b = powf(b, m_HdrGamma);
			m_pData[3*facePixels + xi + m_FaceWidth*yi].set( r, g, b );
		}
	}

	// forward face (negative z)

	for (int yi=0; yi<m_FaceHeight; yi++)
	{
		for (int xi=0; xi<m_FaceWidth; xi++)
		{
			unsigned char *src = getPixelAddr(rawData, m_FaceWidth+xi, 2*m_FaceHeight+yi);
			float r, g, b;
			rgbe2float(&r, &g, &b, src);
			r = powf(r, m_HdrGamma);
			g = powf(g, m_HdrGamma);
			b = powf(b, m_HdrGamma);
			m_pData[4*facePixels + xi + m_FaceWidth*yi].set( r, g, b );
		}
	}

	// backward face (positive z)

	for (int yi=0; yi<m_FaceHeight; yi++)
	{
		for (int xi=0; xi<m_FaceWidth; xi++)
		{
			//unsigned char *src = getPixelAddr(rawData, 2*m_FaceWidth-1-xi, m_FaceHeight-1-yi);
			unsigned char *src = getPixelAddr(rawData, m_FaceWidth+xi, yi);
			float r, g, b;
			rgbe2float(&r, &g, &b, src);
			r = powf(r, m_HdrGamma);
			g = powf(g, m_HdrGamma);
			b = powf(b, m_HdrGamma);
			m_pData[5*facePixels + xi + m_FaceWidth*yi].set( r, g, b );
		}
	}

	delete [] rawData;

	cerr << __FUNCTION__ << ": " << filename << " loaded" << endl;

	return true;
}

void EnvironmentMap::drawSkyBox() const
{
	if ( ! m_SkyDomeTextures[0])
		return;

	glPushAttrib(GL_ENABLE_BIT);
	glDisable(GL_DEPTH_TEST);
	glDisable(GL_LIGHTING);
	glDisable(GL_BLEND);

	glDepthMask(GL_FALSE);

	glEnable(GL_TEXTURE_2D);

#if 1
	float M[16];
	glGetFloatv(GL_MODELVIEW_MATRIX, M);

	// disable translation
	{
		M[12] = M[13] = M[14] = 0.f;
	}

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();
	glMultMatrixf(M);
#endif

	const float boxSize = 20.f;

	glScalef(boxSize, boxSize, boxSize);

	glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);

	// left face (negative x)

	glBindTexture(GL_TEXTURE_2D, m_SkyDomeTextures[0]);
	glBegin(GL_QUADS);
	glTexCoord2f(0.f, 0.f);	glVertex3f(-1.f, -1.f,  1.f);
	glTexCoord2f(1.f, 0.f);	glVertex3f(-1.f, -1.f, -1.f);
	glTexCoord2f(1.f, 1.f);	glVertex3f(-1.f,  1.f, -1.f);
	glTexCoord2f(0.f, 1.f);	glVertex3f(-1.f,  1.f,  1.f);
	glEnd();

	// right face (positive x)

	glBindTexture(GL_TEXTURE_2D, m_SkyDomeTextures[1]);
	glBegin(GL_QUADS);
	glTexCoord2f(0.f, 0.f);	glVertex3f( 1.f, -1.f, -1.f);
	glTexCoord2f(1.f, 0.f);	glVertex3f( 1.f, -1.f,  1.f);
	glTexCoord2f(1.f, 1.f);	glVertex3f( 1.f,  1.f,  1.f);
	glTexCoord2f(0.f, 1.f);	glVertex3f( 1.f,  1.f, -1.f);
	glEnd();

	// top face (positive y)

	glBindTexture(GL_TEXTURE_2D, m_SkyDomeTextures[2]);
	glBegin(GL_QUADS);
	glTexCoord2f(0.f, 0.f);	glVertex3f(-1.f,  1.f, -1.f);
	glTexCoord2f(1.f, 0.f);	glVertex3f( 1.f,  1.f, -1.f);
	glTexCoord2f(1.f, 1.f);	glVertex3f( 1.f,  1.f,  1.f);
	glTexCoord2f(0.f, 1.f);	glVertex3f(-1.f,  1.f,  1.f);
	glEnd();

	// bottom face (negative y)

	glBindTexture(GL_TEXTURE_2D, m_SkyDomeTextures[3]);
	glBegin(GL_QUADS);
	glTexCoord2f(0.f, 0.f);	glVertex3f(-1.f, -1.f,  1.f);
	glTexCoord2f(1.f, 0.f);	glVertex3f( 1.f, -1.f,  1.f);
	glTexCoord2f(1.f, 1.f);	glVertex3f( 1.f, -1.f, -1.f);
	glTexCoord2f(0.f, 1.f);	glVertex3f(-1.f, -1.f, -1.f);
	glEnd();

	// forward face (negative z)

	glBindTexture(GL_TEXTURE_2D, m_SkyDomeTextures[4]);
	glBegin(GL_QUADS);
	glTexCoord2f(0.f, 0.f);	glVertex3f(-1.f, -1.f, -1.f);
	glTexCoord2f(1.f, 0.f);	glVertex3f( 1.f, -1.f, -1.f);
	glTexCoord2f(1.f, 1.f);	glVertex3f( 1.f,  1.f, -1.f);
	glTexCoord2f(0.f, 1.f);	glVertex3f(-1.f,  1.f, -1.f);
	glEnd();

	// backward face (positive z)

	glBindTexture(GL_TEXTURE_2D, m_SkyDomeTextures[5]);
	glBegin(GL_QUADS);
	glTexCoord2f(0.f, 0.f);	glVertex3f(-1.f,  1.f,  1.f);
	glTexCoord2f(1.f, 0.f);	glVertex3f( 1.f,  1.f,  1.f);
	glTexCoord2f(1.f, 1.f);	glVertex3f( 1.f, -1.f,  1.f);
	glTexCoord2f(0.f, 1.f);	glVertex3f(-1.f, -1.f,  1.f);
	glEnd();

	glBindTexture(GL_TEXTURE_2D, 0);

	glPopMatrix();

	glDepthMask(GL_TRUE);

	glPopAttrib();
}

#ifndef GL_CLAMP_TO_EDGE
#define GL_CLAMP_TO_EDGE	0x812F
#endif

void EnvironmentMap::makeGLTextures()
{
	if ( ! m_SkyDomeTextures[0])
		glGenTextures(6, m_SkyDomeTextures);

	glPixelStorei(GL_UNPACK_ALIGNMENT, 1);

	const int facePixels = m_FaceWidth * m_FaceHeight;

	for (int fi=0; fi<6; fi++)
	{
		//cerr << __FUNCTION__ << ": m_SkyDomeTextures[" << fi << "] = " << m_SkyDomeTextures[fi] << endl;
		glBindTexture(GL_TEXTURE_2D, m_SkyDomeTextures[fi]);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
		glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, m_FaceWidth, m_FaceHeight, 0, GL_RGB, GL_FLOAT, &m_pData[fi*facePixels]);
	}

	glBindTexture(GL_TEXTURE_2D, 0);
}

