#pragma once

#include <iostream>
#include <fstream>

#include <clocale>

#include <IL/il.h>
#include <IL/ilu.h>
#include <IL/ilut.h>

#include "ImageRect.h"
#include "data_path.h"

class ImageIO
{
public:
	static void Init()
	{
		ilInit();
		iluInit();
		ilutInit();
		ilutRenderer( ILUT_OPENGL );

		image_path.path.push_back("..");
		image_path.path.push_back("../..");
		image_path.path.push_back("../../..");
		image_path.path.push_back("images");
		image_path.path.push_back("../images");
		image_path.path.push_back("../../images");
	}

	static data_path image_path;

	ImageIO()
	{
	}

	~ImageIO()
	{
	}

	#define PrintILErrorMacro() {		\
		ILenum err_code = ilGetError();	\
										\
		if (err_code != IL_NO_ERROR) {						\
			const ILstring err_str = iluErrorString(err_code);	\
			std::wcerr << "DevIL Error: " << err_str << " (File: "	\
				<< __FILE__ << ", line " << __LINE__ << ")" << std::endl;    \
		}									\
	}

	template <class T>
	bool load(const char *filename, ImageRect<T> &img)
	{
		PrintILErrorMacro();

		ILuint imgName;
		ilGenImages(1, &imgName);
		ilBindImage(imgName);

		ilEnable( IL_ORIGIN_SET );
		ilOriginFunc(IL_ORIGIN_LOWER_LEFT);

		// convert to wchar_t*

		const std::string filePath = image_path.get_file(filename);
		size_t strLength = strlen( filePath.c_str() );
		wchar_t *wc = new wchar_t[ strLength+1 ];

		strLength = mbstowcs( wc, filePath.c_str(), strLength+1 );

		if ( strLength == -1 )
		{
			std::cerr << __FUNCTION__ << ": cannot convert wchar" << std::endl;
		}

		if ( ! ilLoadImage( wc ) )
		{
			std::cerr << __FUNCTION__ << ": cannot open file: " << filename << std::endl;
			PrintILErrorMacro();
			ilDeleteImages(1, &imgName);
			return false;
		}

		delete [] wc;

		//setDevILImageFormat(img);
		ilConvertImage( getILImageFormat(img), getILImageType(img) );
		ilDisable( IL_ORIGIN_SET );

		const int w = ilGetInteger( IL_IMAGE_WIDTH );
		const int h = ilGetInteger( IL_IMAGE_HEIGHT );

		img.allocate( w, h, (T *)ilGetData());

		ilDeleteImages(1, &imgName);

		std::cerr << __FUNCTION__ << ": file loaded: " << filename 
			<< " (" << w << "x" << h << ")" << std::endl;

		PrintILErrorMacro();

		return true;
	}

	template <class T>
	bool save(const char *filename, const ImageRect<T> &img)
	{
		PrintILErrorMacro();

		ILuint imgName;
		ilGenImages(1, &imgName);
		ilBindImage(imgName);

		ImageRect<T> tmp = img;

		ilTexImage(tmp.getWidth(), tmp.getHeight(), 1, getNumChannels(img), getILImageFormat(img), getILImageType(img), tmp.getData());

		// Caution: ilSaveImage() seems to require IL_UNSIGNED_BYTE.
		ilConvertImage( getILImageFormat(img), IL_UNSIGNED_BYTE );

		ilEnable(IL_FILE_OVERWRITE);

		// convert to wchar_t*

		size_t strLength = strlen( filename );
		wchar_t *wc = new wchar_t[ strLength+1 ];

		strLength = mbstowcs( wc, filename, strLength+1 );

		if ( strLength == -1 )
		{
			std::cerr << __FUNCTION__ << ": cannot convert wchar" << std::endl;
		}

		if ( ! ilSaveImage( wc ) )
		{
			std::cerr << __FUNCTION__ << ": cannot save file: " << filename << std::endl;
			PrintILErrorMacro();
			ilDeleteImages(1, &imgName);
			return false;
		}

		delete [] wc;

		ilDisable(IL_FILE_OVERWRITE);

		ilDeleteImages(1, &imgName);

		std::cerr << __FUNCTION__ << ": file saved: " << filename << std::endl;

		PrintILErrorMacro();

		return true;
	}

#if 0
	// Ƃ肠 ppm p 
	template <class T>
	bool save(const char *filename, ImageRect<T> &img)
	{
		std::cerr << __FUNCTION__ << ": not supported" << std::endl;
		return false;
	}

	template <>
	bool save(const char *filename, ImageRect<MyAlgebra::vec3> &img)
	{
		const int w = img.getWidth();
		const int h = img.getHeight();

		if (w <= 0 || h <= 0)
		{
			std::cerr << __FUNCTION__ << ": invalid image size" << std::endl;
			return false;
		}

		std::ofstream ofs(filename, std::ios::binary);

		if ( !ofs)
		{
			std::cerr << __FUNCTION__ << ": cannot open file: " << filename << std::endl;
			return false;
		}

		unsigned char *buf = new unsigned char[3*w*h];
		
		//for (int yi=h-1; yi>=0; yi--)
		for (int yi=0; yi<h; yi++)
		{
			for (int xi=0; xi<w; xi++)
			{
				//const int idx = 3*(xi + w*yi);
				const int idx = 3*(xi + w*(h - 1 - yi));

				buf[idx] = unsigned char(img(xi,yi).r * 255);
				buf[idx+1] = unsigned char(img(xi,yi).g * 255);
				buf[idx+2] = unsigned char(img(xi,yi).b * 255);
			}
		}

		ofs << "P6" << std::endl;
		ofs << w << std::endl;
		ofs << h << std::endl;
		ofs << "255" << std::endl;
		ofs.write((char *)buf, 3*w*h);

		delete [] buf;

		std::cerr << __FUNCTION__ << ": " << filename << " saved" << std::endl;

		return true;
	}

	template <>
	bool save(const char *filename, ImageRect<MyAlgebra::vec4> &img)
	{
		const int w = img.getWidth();
		const int h = img.getHeight();

		if (w <= 0 || h <= 0)
		{
			std::cerr << __FUNCTION__ << ": invalid image size" << std::endl;
			return false;
		}

		std::ofstream ofs(filename, std::ios::binary);

		if ( !ofs)
		{
			std::cerr << __FUNCTION__ << ": cannot open file: " << filename << std::endl;
			return false;
		}

		unsigned char *buf = new unsigned char[3*w*h];
		
		//for (int yi=h-1; yi>=0; yi--)
		for (int yi=0; yi<h; yi++)
		{
			for (int xi=0; xi<w; xi++)
			{
				//const int idx = 3*(xi + w*yi);
				const int idx = 3*(xi + w*(h - 1 - yi));

				buf[idx] = unsigned char(img(xi,yi).r * 255);
				buf[idx+1] = unsigned char(img(xi,yi).g * 255);
				buf[idx+2] = unsigned char(img(xi,yi).b * 255);
			}
		}

		ofs << "P6" << std::endl;
		ofs << w << std::endl;
		ofs << h << std::endl;
		ofs << "255" << std::endl;
		ofs.write((char *)buf, 3*w*h);

		delete [] buf;

		std::cerr << __FUNCTION__ << ": " << filename << " saved" << std::endl;

		return true;
	}
#endif

#undef 	PrintILErrorMacro

private:
	template <class T> ILenum getILImageFormat(const ImageRect<T> &img) const { return IL_RGBA; }
	template <> ILenum getILImageFormat(const ImageRect<UbyteLib::ubyte> &img) const { return IL_LUMINANCE; }
	template <> ILenum getILImageFormat(const ImageRect<UbyteLib::ubyte3> &img) const { return IL_RGB; }
	template <> ILenum getILImageFormat(const ImageRect<UbyteLib::ubyte4> &img) const { return IL_RGBA; }
	template <> ILenum getILImageFormat(const ImageRect<float> &img) const { return IL_LUMINANCE; }
	template <> ILenum getILImageFormat(const ImageRect<MyAlgebra::vec3> &img) const { return IL_RGB; }
	template <> ILenum getILImageFormat(const ImageRect<MyAlgebra::vec4> &img) const { return IL_RGBA; }

	template <class T> ILenum getILImageType(const ImageRect<T> &img) const { return IL_UNSIGNED_BYTE; }
	template <> ILenum getILImageType(const ImageRect<UbyteLib::ubyte> &img) const { return IL_UNSIGNED_BYTE; }
	template <> ILenum getILImageType(const ImageRect<UbyteLib::ubyte3> &img) const { return IL_UNSIGNED_BYTE; }
	template <> ILenum getILImageType(const ImageRect<UbyteLib::ubyte4> &img) const { return IL_UNSIGNED_BYTE; }
	template <> ILenum getILImageType(const ImageRect<float> &img) const { return IL_FLOAT; }
	template <> ILenum getILImageType(const ImageRect<MyAlgebra::vec3> &img) const { return IL_FLOAT; }
	template <> ILenum getILImageType(const ImageRect<MyAlgebra::vec4> &img) const { return IL_FLOAT; }

	template <class T> int getNumChannels(const ImageRect<T> &img) const { return 1; }
	template <> int getNumChannels(const ImageRect<UbyteLib::ubyte> &img) const { return 1; }
	template <> int getNumChannels(const ImageRect<UbyteLib::ubyte3> &img) const { return 3; }
	template <> int getNumChannels(const ImageRect<UbyteLib::ubyte4> &img) const { return 4; }
	template <> int getNumChannels(const ImageRect<float> &img) const { return 1; }
	template <> int getNumChannels(const ImageRect<MyAlgebra::vec3> &img) const { return 3; }
	template <> int getNumChannels(const ImageRect<MyAlgebra::vec4> &img) const { return 4; }

};


