#include "MeshLoader.h"
#include "PlyReader.h"
#include "ObjMesh.h"
#include "ObjLoader.h"
#include "../data_path.h"
#include "../stopwatch.h"

using namespace std;
using namespace MyAlgebra;

bool MeshLoader::load(const char *filename, TriangleMesh &mesh, bool positionArrangementEnabled, bool sizeNormalizationEnabled)
{
	if (filename == "")
	{
		return false;
	}

	const string ext = data_path::getExtension(filename);

	data_path filePath;
	filePath.path.push_back("..");
	filePath.path.push_back("../..");
	filePath.path.push_back("../../..");
	filePath.path.push_back("Meshes");
	filePath.path.push_back("../Meshes");
	filePath.path.push_back("../../Meshes");

	bool isSuccessed = false;

	Stopwatch watch; watch.Start();

	if (ext == "obj")
	{
		isSuccessed = loadObj(filePath.get_file(filename).c_str(), mesh);
	}
	else if (ext == "ply")
	{
		isSuccessed = loadPly(filePath.get_file(filename).c_str(), mesh);
	}

	if (isSuccessed)
	{
		mesh.computeBoundingBox();

		if (positionArrangementEnabled)
		{
			cerr << __FUNCTION__ << ": note: arranging mesh position" << endl;

			const vec3 bboxMin = mesh.getBoundingBoxMin();
			const vec3 bboxMax = mesh.getBoundingBoxMax();

			vector<Triangle> &triangles = mesh.getTriangles();

			for (int tri=0; tri<(int)triangles.size(); tri++)
			{
				triangles[tri].setVertex0( triangles[tri].getVertex0() - bboxMin );
				triangles[tri].setVertex1( triangles[tri].getVertex1() - bboxMin );
				triangles[tri].setVertex2( triangles[tri].getVertex2() - bboxMin );
			}

			mesh.setBoundingBoxMin( vec3(0,0,0) );
			mesh.setBoundingBoxMax( bboxMax - bboxMin );
		}

		if (sizeNormalizationEnabled)
		{
			cerr << __FUNCTION__ << ": note: normalizing mesh size" << endl;

			const vec3 bboxMin = mesh.getBoundingBoxMin();
			const vec3 bboxMax = mesh.getBoundingBoxMax();

			const vec3 scales = bboxMax - bboxMin;

			const float scale = max(scales.x, max(scales.y, scales.z));
			const float invScale = 1.f / scale;

			vector<Triangle> &triangles = mesh.getTriangles();

			for (int tri=0; tri<(int)triangles.size(); tri++)
			{
				triangles[tri].setVertex0( invScale * (triangles[tri].getVertex0() - bboxMin) + bboxMin );
				triangles[tri].setVertex1( invScale * (triangles[tri].getVertex1() - bboxMin) + bboxMin );
				triangles[tri].setVertex2( invScale * (triangles[tri].getVertex2() - bboxMin) + bboxMin );
			}
			
			mesh.setBoundingBoxMin( bboxMin );
			mesh.setBoundingBoxMax( invScale * (bboxMax - bboxMin) + bboxMin );
		}

		cerr << __FUNCTION__ << ": " << filename << " loaded ("
			 << mesh.getNumTriangles() << " tris, " << watch.GetTime() << " sec)" << endl;
		return true;
	}

	cerr << __FUNCTION__ << ": " << filename << " cannot loaded" << endl;

	return false;
}

// private functions -----------------------------------------------------

bool MeshLoader::loadObj(const char *filename, TriangleMesh &mesh)
{
	ObjMesh objMesh;
	ObjLoader<ObjMesh> loader;

	if ( ! loader.load(&objMesh, filename))
	{
		return false;
	}

	const vector<vec3> &vertices = objMesh.getVertices();

	const int nTriangles = objMesh.getNumFaces();

	vector<Triangle> &triangles = mesh.getTriangles();
	triangles.resize( nTriangles );

	for (int tri=0; tri<nTriangles; tri++)
	{
		const vector<int> &faceVertices = objMesh.getFaceVertices(tri);

		triangles[tri].setVertex0( vertices[ faceVertices[0] ] );
		triangles[tri].setVertex1( vertices[ faceVertices[1] ] );
		triangles[tri].setVertex2( vertices[ faceVertices[2] ] );
	}

	const vector<vec3> &normals = objMesh.getNormals();

	if (mesh.getShadingType() == TriangleMesh::Smooth_Shading)
	{
		if ( ! normals.empty() )
		{
			for (int tri=0; tri<nTriangles; tri++)
			{
				const vector<int> &faceVertices = objMesh.getFaceVertices(tri);

				triangles[tri].setNormal0( normals[ faceVertices[0] ] );
				triangles[tri].setNormal1( normals[ faceVertices[1] ] );
				triangles[tri].setNormal2( normals[ faceVertices[2] ] );
			}
		}
		else
		{
			const int nVerts = (int)vertices.size();
			vec3 *tmpNormals = new vec3[ nVerts ];

			unsigned int *faceIndices = new unsigned int[ 3*nTriangles ];
			for (int tri=0; tri<nTriangles; tri++)
			{
				const vector<int> &faceVertices = objMesh.getFaceVertices(tri);

				faceIndices[3*tri  ] = faceVertices[0];
				faceIndices[3*tri+1] = faceVertices[1];
				faceIndices[3*tri+2] = faceVertices[2];
			}

			computeSmoothNormals(nVerts, &vertices[0], tmpNormals, nTriangles, faceIndices);

			for (int tri=0; tri<nTriangles; tri++)
			{
				triangles[tri].setNormal0( tmpNormals[ faceIndices[3*tri  ] ] );
				triangles[tri].setNormal1( tmpNormals[ faceIndices[3*tri+1] ] );
				triangles[tri].setNormal2( tmpNormals[ faceIndices[3*tri+2] ] );
			}

			delete [] tmpNormals;
			delete [] faceIndices;
		}
	}
	else	// mesh.getShadingType() == TriangleMesh::Flat_Shading
	{
		for (int tri=0; tri<nTriangles; tri++)
		{
			const vec3 &v0 = triangles[tri].getVertex0();
			const vec3 &v1 = triangles[tri].getVertex1();
			const vec3 &v2 = triangles[tri].getVertex2();

			const vec3 normal = normalize( (v1-v0)^(v2-v0) );

			triangles[tri].setNormal0( normal );
			triangles[tri].setNormal1( normal );
			triangles[tri].setNormal2( normal );
		}
	}

	return true;
}

bool MeshLoader::loadPly(const char *filename, TriangleMesh &mesh)
{
	if ( ! PlyReader::read(filename) )
		return false;

	vector<Triangle> &triangles = mesh.getTriangles();

	const int nTriangles = PlyReader::getNumTriangles();
	const unsigned int *faceIndices = PlyReader::getIndices();

	triangles.resize( nTriangles );

	const int nVerts = PlyReader::getNumVertices();	
	const vec3 *vertices = PlyReader::getVertices();

	for (int tri=0; tri<nTriangles; tri++)
	{
		triangles[tri].setVertex0( vertices[ faceIndices[3*tri  ] ] );
		triangles[tri].setVertex1( vertices[ faceIndices[3*tri+1] ] );
		triangles[tri].setVertex2( vertices[ faceIndices[3*tri+2] ] );
	}

	if (mesh.getShadingType() == TriangleMesh::Smooth_Shading)
	{
		if (PlyReader::getNormals() != 0)
		{
			const vec3 *normals = PlyReader::getNormals();

			for (int tri=0; tri<nTriangles; tri++)
			{
				triangles[tri].setNormal0( normals[ faceIndices[3*tri  ] ] );
				triangles[tri].setNormal1( normals[ faceIndices[3*tri+1] ] );
				triangles[tri].setNormal2( normals[ faceIndices[3*tri+2] ] );
			}
		}
		else
		{
			vec3 *normals = new vec3[ nVerts ];

			computeSmoothNormals(nVerts, vertices, normals, nTriangles, faceIndices);

			for (int tri=0; tri<nTriangles; tri++)
			{
				triangles[tri].setNormal0( normals[ faceIndices[3*tri  ] ] );
				triangles[tri].setNormal1( normals[ faceIndices[3*tri+1] ] );
				triangles[tri].setNormal2( normals[ faceIndices[3*tri+2] ] );
			}

			delete [] normals;
		}
	}
	else	// mesh.getShadingType() == TriangleMesh::Flat_Shading
	{
		for (int tri=0; tri<nTriangles; tri++)
		{
			const vec3 &v0 = triangles[tri].getVertex0();
			const vec3 &v1 = triangles[tri].getVertex1();
			const vec3 &v2 = triangles[tri].getVertex2();

			const vec3 normal = normalize( (v1-v0)^(v2-v0) );

			triangles[tri].setNormal0( normal );
			triangles[tri].setNormal1( normal );
			triangles[tri].setNormal2( normal );
		}
	}

	return true;
}

void MeshLoader::computeSmoothNormals(int nVerts, const vec3 *verts, vec3 *normals, int nTris, const unsigned int *tris) const
{
	vec3 *faceNormals = new vec3[ nTris ];
	vector<int> *vertexFace = new vector<int>[ nTris ];

	// computing face normals and areas
	for (int tri=0; tri<nTris; tri++)
	{
		// TODO: fill here
	}

	// computing vertex normals by averaging face normals
	for (int v=0; v<nVerts; v++)
	{
		// TODO: fill here
	}

	delete [] faceNormals;
	delete [] vertexFace;
}

