#pragma once

#include <string>
#include <fstream>
#include <sstream>
#include <iostream>
#include <vector>
#include "../my_algebra.h"

// See the following URL for MTL format
// http://www.csit.fsu.edu/~burkardt/data/mtl/mtl.html

template< class M >
class ObjLoader
{
public:
	typedef M Mesh;

	ObjLoader()
	{
	}

	~ObjLoader()
	{
	}

	bool load( Mesh *pMesh, const char * filename)
	{
		const std::string ext = getExtension(filename);

		if ("obj" != ext)
		{
			std::cerr << "ObjLoader::loader: error: invalid extention: " << ext << std::endl;
			return false;
		}

		std::ifstream ifp( filename );
		if ( ! ifp)
		{
			std::cerr << "ObjLoader::loader: error: cannot open file: " << filename << std::endl;
			return false;
		}

		while ( ! ifp.eof() )
		{
			std::string line;
			std::getline( ifp, line );
			std::istringstream is(line);

			if (line[0] == '#') continue;	// skip comment
			if (line == "") continue;

			std::string entryName = getToken( is );
			
			if (entryName == "g")
			{
				readGroupName( pMesh, is );
			}
			else if (entryName == "vn")
			{
				readNormal( pMesh, is );
			}
			else if (entryName == "vt")
			{
				readTexcoord( pMesh, is );
			}
			else if (entryName == "v")
			{
				readVertex( pMesh, is );
			}
			else if (entryName == "f")
			{
				readFace( pMesh, is );
			}
			else if (entryName == "usemtl")
			{
				readMaterial( pMesh, is );
			}
			else if (entryName == "mtllib")
			{
				readMaterialFile( pMesh, is );
			}
		}

		ifp.close();

		return true;
	}

private:
	std::string getExtension(const std::string &filename)
	{
		std::string::size_type t = filename.rfind('.');
		return (t == std::string::npos) ? "" : filename.substr(t+1);
	}

	void readGroupName( Mesh *pMesh, std::istringstream &is )
	{
		std::string groupName;
		is >> groupName;
		pMesh->setGroupName( groupName );
	}

	void readMaterial( Mesh *pMesh, std::istringstream &is )
	{
		std::string materialName;
		is >> materialName;
		pMesh->setMaterialName( materialName );
	}

	void readMaterialFile( Mesh *pMesh, std::istringstream &is )
	{
		std::string materialFile;
		is >> materialFile;

		std::ifstream imfp( materialFile.c_str() );

		if ( ! imfp)
		{
			std::cerr << "ObjLoader::readMaterialFile: cannot open file: "
				<< materialFile << std::endl;
			return;
		}
		
		bool firstMaterialEnded = false;

		while ( ! imfp.eof() )
		{
			std::string line;
			std::getline( imfp, line );
			std::istringstream ims(line);

			if (line[0] == '#') continue;	// skip comment
			if (line == "") continue;

			std::string entryName = getToken( ims );

			//std::cerr << "readMaterialFile: token = " << entryName << std::endl;

			if (entryName == "newmtl")
			{
				if ( ! firstMaterialEnded)
					firstMaterialEnded = true;
				else
					break;

				std::string name;
				ims >> name;
				pMesh->setMaterialName( name );
			}
			else if (entryName == "Ka")
			{
				M::vec3 c;
				ims >> c.r >> c.g >> c.b;
				pMesh->setAmbient( c );
			}
			else if (entryName == "Kd")
			{
				M::vec3 c;
				ims >> c.r >> c.g >> c.b;
				pMesh->setDiffuse( c );
			}
			else if (entryName == "Ks")
			{
				M::vec3 c;
				ims >> c.r >> c.g >> c.b;
				pMesh->setSpecular( c );
			}
			else if (entryName == "d")
			{
				float alpha;
				ims >> alpha;
				pMesh->setAlpha( alpha );
			}
			else if (entryName == "Ns")
			{
				float shininess;
				ims >> shininess;
				pMesh->setShininess( shininess );
			}
			else if (entryName == "map_Kd")
			{
				std::string name;
				ims >> name;
				pMesh->setTexName( name );
			}
		}

		imfp.close();
	}

	void readTexcoord( Mesh *pMesh, std::istringstream &is )
	{
		M::vec2 uv;
		is >> uv.s >> uv.t;
		pMesh->addTexcoord( uv );
	}

	void readNormal( Mesh *pMesh, std::istringstream &is )
	{
		M::vec3 n;
		is >> n.x >> n.y >> n.z;
		pMesh->addNormal( n );
	}

	void readVertex( Mesh *pMesh, std::istringstream &is )
	{
		M::vec3 p;
		is >> p.x >> p.y >> p.z;
		pMesh->addVertex( p );
	}

	void readFace( Mesh *pMesh, std::istringstream &is )
	{
		std::vector<int> faceVertex;
		std::vector<int> faceTexture;
		std::vector<int> faceNormal;

		while (! is.eof())
		{
			int vert_idx;
			int tex_idx;
			int norm_idx;

			if (! getInt( &vert_idx, is ) )
			{
				std::cerr << "ObjLoader::readFace: invalid vertex index" << std::endl;
				return;
			}

			faceVertex.push_back( vert_idx-1 );

			if (! getSlash( is ) ) continue;

			if ( getInt( &tex_idx, is ) )
	            faceTexture.push_back( tex_idx-1 );
			
			if (! getSlash( is ) ) continue;

			if (isspace(is.get()))
			{
				skipSpace( is );
				skipComment( is );
				std::cerr << "3rd continue" << std::endl;
				continue;
			}
			is.unget();

			if ( getInt( &norm_idx, is ) )
				faceNormal.push_back( norm_idx-1 );
			
			skipSpace( is );
			skipComment( is );
		}

		pMesh->addFaceVertex( faceVertex );
		if (! faceTexture.empty() ) pMesh->addFaceTexcoord( faceTexture );
		if (! faceNormal.empty() ) pMesh->addFaceNormal( faceNormal );
	}
	
	std::string getToken( std::istringstream &is )
	{
		std::string token;
		is >> token;
		return token;
	}

	bool getInt( int *pIdx, std::istringstream &is )
	{
		skipSpace( is );

		std::string digit;

		int length = is.tellg();

		char c = is.get();
		if (! isdigit(c) )
		{
			is.unget();
			//std::cerr << "ObjLoader::getInt: non digit: " << c << std::endl;
			return false;
		}

		while (isdigit(c))
		{
			digit.push_back( c );
			c = is.get();
		}

		if (c == '.')
		{
			is.seekg(length);
			//std::cerr << "ObjLoader::getInt: floating point not allowed" << std::endl;
			return false;
		}

		is.unget();

		*pIdx = atoi(digit.c_str());

		return true;
	}

	bool getSlash( std::istringstream &is )
	{
		skipSpace( is );

		char c = is.get();
		if (c == '/') return true;
		is.unget();
		return false;
	}

	void skipSpace( std::istringstream &is )
	{
		char c = is.get();
		while (isspace(c)) c = is.get();
		is.unget();
	}

	void skipComment( std::istringstream &is )
	{
		char c = is.get();
		if (c == '#')
			is.seekg(ios::end);
		else
			is.unget();
	}
};

