#include <cstdlib>
#include <fstream>
#include "RecursiveRayTracer.h"
#include "Camera.h"
#include "Scene.h"
#include "ImageRect.h"
#include "EnvironmentMap.h"
#include "OGLDisplayDriver.h"
#include "PrintGLErrorMacro.h"
#include "DebugIntersectionLog.h"

//#include "SceneLoader.h"
#include "ConfigLoader.h"

//#define USE_WINUTIL

#include "ImageIO.h"
#include "WinUtil.h"

#include "CameraViewWindow.h"
#include "WorldViewWindow.h"

#include "MiscSettings.h"

using namespace std;
using namespace MyAlgebra;

extern RayTracer *g_pRayTracer;

extern ImageRect<vec3> g_ScreenBuffer;
extern Camera g_Camera;
extern Scene g_Scene;

extern DebugIntersectionLog debugLog;
extern vector<vec3> debug_hit_points;

bool g_SingleRayDebug = false;
vector<int> g_TracedGridCells;
vec3 g_DebugHitPoint;

extern MiscSettings g_MiscSettings;

bool g_ScreenUpdateRequired = false;
bool g_DisplayRayTracingResult = false;
int g_KeepRayTracing = 0;

CameraViewWindow g_CameraViewWindow;
WorldViewWindow g_WorldViewWindow;

int g_MainWindowID = 0;

namespace OGLDisplayDriver 
{
	// private variables and functions

	static GLUI *g_pGLUI = 0;

	static GLUI_StaticText *glui_scene_filename = 0;

	static GLuint screen_tex = 0;

	// for GLUI

	enum GLUI_Callback_ID
	{
		Execute_Ray_Tracing_ID,
		Load_Scene_File_ID,
		Reload_Scene_File_ID,
		Save_Screenshot_ID,
		Toggle_Keep_Ray_Tracing_ID,
		Transform_Updated_ID,
		Dump_Camera_Info_ID,
	};

	static void GLUICallback(int id)
	{
		glutSetWindow(g_MainWindowID);

		switch (id)
		{
		case Execute_Ray_Tracing_ID:
			//cout << "Rendering ... " << endl;

			g_pRayTracer->renderScene( g_ScreenBuffer );

			g_DisplayRayTracingResult = true;
			g_ScreenUpdateRequired = false;

			break;
		case Toggle_Keep_Ray_Tracing_ID:
			if ( ! g_KeepRayTracing)
			{
				g_DisplayRayTracingResult = false;
			}
			else
			{
				g_ScreenUpdateRequired = true;
			}
			break;
		case Load_Scene_File_ID:
			{
#ifdef USE_WINUTIL
				//const string filename = WinUtil::OpenFileDialog("txt", "Scene File(*.txt)\0*.txt\0All Files(*.*)\0*.*\0\0");
				const string filename = WinUtil::OpenFileDialog();

#if 1
				ConfigLoader loader;

				if ( loader.load( filename.c_str(), 1==g_MiscSettings.m_SceneVerboseLoad ) )
				{
#else
				SceneLoader loader;

				if ( loader.load(g_Scene, filename.c_str()) )
				{
#endif
					//scene_filename = filename;
					g_MiscSettings.m_SceneFilename = filename;

					char buf[1024];
					sprintf(buf, "File: %s", data_path::getFilename(filename).c_str());
					glui_scene_filename->set_text( buf );

					g_ScreenUpdateRequired = true;
				}
#else
				ConfigLoader loader;

				if ( loader.load( g_MiscSettings.m_SceneFilename.c_str(), 1==g_MiscSettings.m_SceneVerboseLoad ) )
				{
					g_ScreenUpdateRequired = true;
				}
#endif
			}
			break;
		case Reload_Scene_File_ID:
			{
#if 1
				ConfigLoader loader;

				if ( loader.load( g_MiscSettings.m_SceneFilename.c_str(), 1==g_MiscSettings.m_SceneVerboseLoad ) )
				{
#else
				SceneLoader loader;

				if ( loader.load(g_Scene, g_MiscSettings.m_SceneFilename.c_str()) )
				{
#endif
					g_ScreenUpdateRequired = true;
				}
			}
			break;
		case Save_Screenshot_ID:
			{
#ifdef USE_WINUTIL
				const string filename = WinUtil::SaveFileDialog("jpg");

				ImageIO io;

				io.save(filename.c_str(), g_ScreenBuffer);
#else
				ImageIO io;

				io.save(g_MiscSettings.m_ScreenshotFilename.c_str(), g_ScreenBuffer);
#endif
			}
			break;
		case Dump_Camera_Info_ID:
			{
				vec3 position, center, updir;
				g_Camera.getLookAt(position, center, updir);

				cout << "camera: " << endl
					 << "  position = " << position << ";" << endl 
					 << "  center = " << center << ";" << endl 
					 << "  updir = " << updir << ";" << endl;

				mat4 M;
				g_Camera.getModelviewMatrix( M.mat_array );

				cout << "modelview: " << M << endl;
			}
			break;
		default:
			break;
		}

		g_pGLUI->sync_live();
		glutPostRedisplay();
	}

	// public functions

	void Init(int argc, char **argv)
	{
		glutInit(&argc, argv);
		glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE | GLUT_DEPTH );

		int xWindowStartPos = 50;
		int yWindowStartPos = 50;
		const int xWindowOffset = 30;
		const int yWindowOffset = 50;

		g_CameraViewWindow.setInitialWindowSize(g_ScreenBuffer.getWidth(), g_ScreenBuffer.getHeight());
		g_CameraViewWindow.setInitialWindowPos(xWindowStartPos, yWindowStartPos);
		g_CameraViewWindow.create("Rendered Result");

		g_CameraViewWindow.setImageBuffer( &g_ScreenBuffer );

	
		g_MainWindowID = g_CameraViewWindow.getWindowID();


		xWindowStartPos += xWindowOffset+g_ScreenBuffer.getWidth();

		g_WorldViewWindow.setInitialWindowSize(g_ScreenBuffer.getWidth(), g_ScreenBuffer.getHeight());
		g_WorldViewWindow.setInitialWindowPos(xWindowStartPos, yWindowStartPos);
		g_WorldViewWindow.create("World View");

		xWindowStartPos += xWindowOffset+g_ScreenBuffer.getWidth();



		glClearDepth(100000.0);
		//glDepthFunc(GL_LEQUAL);

		glColorMaterial(GL_FRONT, GL_DIFFUSE);
		glEnable(GL_COLOR_MATERIAL);

		glEnable(GL_POINT_SMOOTH);
		glHint(GL_POINT_SMOOTH_HINT, GL_DONT_CARE);

		glEnable(GL_LINE_SMOOTH);
		glHint(GL_LINE_SMOOTH_HINT, GL_DONT_CARE);

		glClearColor(1,1,1,1);

		// initialize transformation

		ImageIO::Init();

		// initialize GLUI

		// ray trace

		g_pGLUI = GLUI_Master.create_glui("GLUI", 0, xWindowStartPos+xWindowOffset, 0);

		{
			GLUI_Panel *rayTracePanel = g_pGLUI->add_panel("Ray Trace");

			g_pGLUI->add_button_to_panel(rayTracePanel, "Execute", Execute_Ray_Tracing_ID, GLUICallback);

			g_pGLUI->add_checkbox_to_panel(rayTracePanel, "Keep Ray Tracing", &g_KeepRayTracing, Toggle_Keep_Ray_Tracing_ID, GLUICallback);
			g_pGLUI->add_checkbox_to_panel(rayTracePanel, "Dump Comp. Time", &g_MiscSettings.m_DumpComputationTime);

			GLUI_Spinner *depthSpinner = g_pGLUI->add_spinner_to_panel(rayTracePanel,
				"Max Recursion Depth", GLUI_SPINNER_INT, &RecursiveRayTracer::max_recursion_depth);
			depthSpinner->set_int_limits(0, 100);
		}

		// scene

		{
			GLUI_Panel *scenePanel = g_pGLUI->add_panel("Scene");

			g_pGLUI->add_button_to_panel(scenePanel, "Load", Load_Scene_File_ID, GLUICallback);

#ifdef USE_WINUTIL
			char buf[1024];
			sprintf(buf, "File: %s", data_path::getFilename(g_MiscSettings.m_SceneFilename).c_str());
			glui_scene_filename = new GLUI_StaticText(scenePanel, buf);
#else
			GLUI_EditText *text = new GLUI_EditText(scenePanel, "Filename: ", g_MiscSettings.m_SceneFilename);
			text->set_w( 175 );
#endif

			g_pGLUI->add_button_to_panel(scenePanel, "Reload", Reload_Scene_File_ID, GLUICallback);

			g_pGLUI->add_checkbox_to_panel(scenePanel, "Verbose", &g_MiscSettings.m_SceneVerboseLoad);
		}

		{
			GLUI_Panel *screenshotPanel = g_pGLUI->add_panel("Screenshot");

#ifndef USE_WINUTIL
			GLUI_EditText *text = new GLUI_EditText(screenshotPanel, "Filename: ", g_MiscSettings.m_ScreenshotFilename);
			text->set_w( 175 );
#endif

			g_pGLUI->add_button_to_panel(screenshotPanel, "Save", Save_Screenshot_ID, GLUICallback);
		}

		// camera

		{
			GLUI_Panel *cameraPanel = g_pGLUI->add_panel("Camera");

			g_pGLUI->add_button_to_panel(cameraPanel, "Dump Info", Dump_Camera_Info_ID, GLUICallback);
		}

		g_pGLUI->add_column();

		{
			GLUI_Panel *p = g_pGLUI->add_panel("World View");

			g_pGLUI->add_checkbox_to_panel(p, "Toggle Wireframe", &WorldViewWindow::s_ToggleWireframe);
			g_pGLUI->add_checkbox_to_panel(p, "Draw Objects", &WorldViewWindow::s_DrawObjects);
			g_pGLUI->add_checkbox_to_panel(p, "Draw View Frustum", &WorldViewWindow::s_DrawViewFrustum);
			g_pGLUI->add_checkbox_to_panel(p, "Draw World Axes", &WorldViewWindow::s_DrawWorldAxes);
		}

		g_pGLUI->add_button("Quit", -1, exit);

		GLUI_Master.set_glutIdleFunc(OGLBaseWindow::Idle);

		g_pGLUI->set_main_gfx_window(g_MainWindowID);
	}

	void MainLoop()
	{
		glutMainLoop();
	}

}
