#include "RecursiveRayTracer.h"
#include "Scene.h"
#include "Camera.h"
#include "Sampler.h"
#include "Filter.h"

#include "PerfectSpecularMaterial.h"
#include "SpecularRefractionMaterial.h"

#include "EnvironmentMap.h"

#include "ImageIO.h"
#include "BinaryImageIO.h"

#include "stopwatch.h"

#include "MiscSettings.h"

using namespace std;
using namespace MyAlgebra;

extern Camera g_Camera;
extern Scene g_Scene;

extern int g_CurrentXi;
extern int g_CurrentYi;

extern bool g_SingleRayDebug;
extern vec3 g_DebugHitPoint;

extern MiscSettings g_MiscSettings;

int RecursiveRayTracer::max_recursion_depth = 5;
int RecursiveRayTracer::s_NumSamplesPerPixel = 1;

const float t_epsilon = 0.01f;
const float t_shadow_ray_epsilon = 0.01f;
const float t_infinity = 1.0e+10f;

void RecursiveRayTracer::renderScene(ImageRGBf &imageBuffer)
{
	Stopwatch watch;

	if (g_MiscSettings.m_DumpComputationTime)
		watch.Start();

	// get axes of camera coordinate system
	vec3 u, v, w;
	g_Camera.getModelviewCoordinates(u, v, w);

	const vec3 eye = g_Camera.getEyePosition();

	const int width = imageBuffer.getWidth();
	const int height = imageBuffer.getHeight();

	const Real screenDist = g_Camera.calcScreenDistance();

	const Real halfWidth = 0.5f * width;
	const Real halfHeight = 0.5f * height;

	const int nSamples = m_PixelSampler->adjustNumSamples(s_NumSamplesPerPixel);

	vec2 *pixelOffsets = new vec2[nSamples];
	vec3 *samples = new vec3[nSamples];

	for (int yi=0; yi<height; yi++)
	{
		g_CurrentYi = yi;	// for debugging

		for (int xi=0; xi<width; xi++)
		{
			g_CurrentXi = xi;	// for debugging

			// assume that the provided samples are within [0, 1]^2
			m_PixelSampler->sample(pixelOffsets, nSamples);

			for (int si=0; si<nSamples; si++)
			{
				// setup viewing ray
				vec3 dir = (xi+pixelOffsets[si].x-halfWidth)*u
					+ (yi+pixelOffsets[si].y-halfHeight)*v
					- screenDist*w;
				dir.normalize();

				samples[si] = trace( Ray(eye, dir) );
			}

			imageBuffer(xi, yi) = m_PixelFilter->filter(samples, pixelOffsets, nSamples);
		}
	}

	delete [] pixelOffsets;
	delete [] samples;

	if (g_MiscSettings.m_DumpComputationTime)
		cout << __FUNCTION__ << ": time: " << watch.GetTime() << " sec" << endl;
}

vec3 RecursiveRayTracer::singleRayTrace(const ImageRGBf &imageBuffer, int xi, int yi )
{
	// get axes of camera coordinate system
	vec3 u, v, w;
	g_Camera.getModelviewCoordinates(u, v, w);

	const vec3 eye = g_Camera.getEyePosition();

	const int width = imageBuffer.getWidth();
	const int height = imageBuffer.getHeight();

	const Real screenDist = g_Camera.calcScreenDistance();

	const Real halfWidth = 0.5f * width;
	const Real halfHeight = 0.5f * height;

	g_CurrentXi = xi;	// for debugging
	g_CurrentYi = yi;	// for debugging

	// setup viewing ray
	vec3 dir = (xi-halfWidth)*u + (yi-halfHeight)*v - screenDist*w;
	dir.normalize();

	return trace( Ray(eye, dir) );
}

vec3 RecursiveRayTracer::traceRec(const Ray &ray, int recursionDepth)
{
	if (recursionDepth > max_recursion_depth)
		return ( g_Scene.isEnvironmentMapAvailable() )
			? g_Scene.getEnvironmentMap()->fetchColor(ray.getUnitDir())
			: g_Scene.getBackgroundColor();

	HitRecord record;
	record.t = t_infinity;

	bool hitObject = false;

	for (int oi=0; oi<g_Scene.getNumObjects(); oi++)
	{
		GeometricObject *o = g_Scene.getObject(oi);

		HitRecord tmpRec;
		const bool isHit = o->hit(ray, t_epsilon, t_infinity, tmpRec);

		if (isHit && record.t > tmpRec.t)
		{
			record = tmpRec;
			hitObject = true;
		}
	}

	if ( g_SingleRayDebug )
	{
		g_DebugHitPoint =  ray.getPosition( (hitObject) ? record.t : t_infinity );
	}

	if ( ! hitObject)
		return ( g_Scene.isEnvironmentMapAvailable() )
			? g_Scene.getEnvironmentMap()->fetchColor(ray.getUnitDir())
			: g_Scene.getBackgroundColor();

	const vec3 hitPos = record.hit_pos;
	Material *material = record.material;

	const Material::Material_Type matType = material->getMaterialType();

	if (matType == Material::Perfect_Specular_Type)
	{
		// TODO: fill here to render mirror-like (perfect specular) materials
		// normal is given via record.normal

		return vec3(0.f);
	}
	else if (matType == Material::Specular_Refraction_Type)
	{
		// TODO: fill here to render glass-like (refractive) materials
		// normal is given via record.normal, and the hit position is given via record.hit_pos

		return vec3(0.f);
	}

	// lights

	vec3 color(0,0,0);

	for (int li=0; li<g_Scene.getNumLights(); li++)
	{
		LightSource *l = g_Scene.getLight(li);

		bool isShadowed = false;

		LightInfo lightInfo;
		l->calcLightInfo(hitPos, lightInfo);

		Ray toLight(hitPos, lightInfo.dir);

		for (int oi=0; oi<g_Scene.getNumObjects(); oi++)
		{
			GeometricObject *o = g_Scene.getObject(oi);
			const bool isHit = o->shadowHit(toLight, t_shadow_ray_epsilon, lightInfo.dist);

			if (isHit)
			{
				isShadowed = true;
				break;
			}
		}

		if (isShadowed) continue;

		vec3 tmpColor = material->calcReflectedLight( record, -lightInfo.dir, -ray.getUnitDir() );
		mult(tmpColor, tmpColor, lightInfo.color);

		color += tmpColor;
	}

	// ambient color is added here
	color += material->getAmbientCoeff();

	return color;
}

