#include "SceneLoader.h"
#include "data_path.h"

#include <fstream>
#include <cstring>
#include <string>

#include "RecursiveRayTracer.h"

#include "Scene.h"

#include "Sphere.h"
#include "Plane.h"
#include "Triangle.h"
#include "TexturedPlane.h"
#include "TexturedSphere.h"
#include "SolidTexturedSphere.h"

#include "PointLight.h"
#include "DirectionalLight.h"

#include "DiffuseMaterial.h"
#include "PhongMaterial.h"
#include "PerfectSpecularMaterial.h"
#include "SpecularRefractionMaterial.h"
#include "TexturedMaterial.h"

#include "CheckeredTexture.h"
#include "SolidCheckeredTexture.h"
#include "NoiseTexture.h"
#include "MarbleTexture.h"
#include "WoodenTexture.h"
#include "TurbulenceTexture.h"
#include "DerivTurbulenceTexture.h"
#include "ImageTexture.h"

#include "BoxFilter.h"

#include "SingleSampler.h"
#include "RandomSampler.h"
#include "Regular2x2Sampler.h"
#include "Regular3x3Sampler.h"
#include "RegularNxNSampler.h"
#include "JitterSampler.h"
#include "MultiJitterSampler.h"

#include "EnvironmentMap.h"

#include "ImageRect.h"

#include "OGLDisplayDriver.h"

using namespace std;
using namespace MyAlgebra;

extern RayTracer *g_pRayTracer;

extern ImageRect<vec3> g_ScreenBuffer;
extern Camera g_Camera;
extern Scene g_Scene;
//extern EnvironmentMap *g_pEnvironmentMap;

//extern int max_depth;

//extern vec3 background_color;

extern int g_MainWindowID;

bool SceneLoader::load(Scene &scene, const char *filename)
{
	data_path scenePath;

	scenePath.path.push_back(".");
	scenePath.path.push_back("..");
	scenePath.path.push_back("../..");

	ifstream ifs( scenePath.get_file(filename).c_str() );

	if ( ! ifs)
	{
		cerr << __FUNCTION__ << ": cannot open file: " << filename << endl;
		return false;
	}

	cerr << __FUNCTION__ << ": loading file: " << filename << endl;

	int nErrors = 0;

	line_number = 0;

	lights.clear();
	objects.clear();

	//if ( ! current_material)
	//{
	//	delete current_material;
	//	current_material = 0;
	//}

	Material::ClearMaterialCache();
	current_material = 0;

	//Texture::ClearTextureCache();

	specified_envmap = 0;
	pixel_sampler = 0;
	pixel_filter = 0;

	// store current setting

	screen_width = g_ScreenBuffer.getWidth();
	screen_height = g_ScreenBuffer.getHeight();
	g_Camera.getLookAt(camera_position, camera_center, camera_up);
	camera_fovy = g_Camera.getFieldOfView();
	//_max_depth = max_depth;
	_max_recursion_depth = RecursiveRayTracer::max_recursion_depth;
	//_background_color = background_color;
	_background_color = scene.getBackgroundColor();
	direct_light_scaling = 1.f;

	while ( ! ifs.eof())
	{
		string line;
		getline(ifs, line);

		line_number++;

		if (line == "" || line[0] == '#')
			continue;

		iss.str(line);

		nErrors += parseLine();

		iss.clear();
	}

	if (nErrors)
	{
		cerr << __FUNCTION__ << ": " << nErrors << " error(s)" << endl;
		cerr << __FUNCTION__ << ": " << filename << " cannot be loaded" << endl;
		return false;
	}

	// update setting

	scene.clearScene();

	for (int i=0; i<(int)lights.size(); i++)
		scene.addLight( lights[i] );

	for (int i=0; i<(int)objects.size(); i++)
		scene.addObject( objects[i] );

	lights.clear();
	objects.clear();

	//if (g_pEnvironmentMap) delete g_pEnvironmentMap;
	//g_pEnvironmentMap = specified_envmap;
	//if (g_pEnvironmentMap) g_pEnvironmentMap->makeGLTextures();
	scene.setEnvironmentMap( specified_envmap );

	g_pRayTracer->registerImageSampler((pixel_sampler) ? pixel_sampler : new SingleSampler());
	g_pRayTracer->registerImageFilter((pixel_filter) ? pixel_filter : new BoxFilter());

	g_ScreenBuffer.allocate(screen_width, screen_height);
	g_Camera.setLookAt(camera_position, camera_center, camera_up);
	g_Camera.setFieldOfView(camera_fovy);
	//max_depth = _max_depth;
	RecursiveRayTracer::max_recursion_depth = _max_recursion_depth;
	//background_color = _background_color;
	scene.setBackgroundColor( _background_color );

	cerr << __FUNCTION__ << ": done" << endl;

	return true;
}

// private functions

bool SceneLoader::parseLine()
{
	error_in_line = false;

	char cmd[256];
	iss >> cmd;

	//cout << __FUNCTION__ << ": cmd = " << cmd << endl;

	if ( ! strcmp("screen_size", cmd))
	{
		error_in_line |= readInteger(screen_width);
		error_in_line |= readInteger(screen_height);
		error_in_line |= requireEol();

		if ( ! error_in_line && g_MainWindowID)
		{
			glutReshapeWindow(screen_width, screen_height);
		}
	}
	else if ( ! strcmp("max_depth", cmd) ||	// for compatibility
			  ! strcmp("max_recursion_depth", cmd))
	{
		error_in_line |= readInteger(_max_recursion_depth);
		error_in_line |= requireEol();
	}
	else if ( ! strcmp("camera_position", cmd))
	{
		error_in_line |= readVector(camera_position);
		error_in_line |= requireEol();
	}
	else if ( ! strcmp("camera_center", cmd))
	{
		error_in_line |= readVector(camera_center);
		error_in_line |= requireEol();
	}
	else if ( ! strcmp("camera_up", cmd))
	{
		error_in_line |= readVector(camera_up);
		error_in_line |= requireEol();
	}
	else if ( ! strcmp("camera_fovy", cmd))
	{
		error_in_line |= readFloat(camera_fovy);
		error_in_line |= requireEol();
	}
	else if ( ! strcmp("pixel_sampler", cmd))
	{
		string samplerType;

		error_in_line |= readString(samplerType);

		Sampler *sampler = 0;
		
		if (samplerType == "single")
		{
			sampler = new SingleSampler();
		}
		else if (samplerType == "regular2x2")
		{
			sampler = new Regular2x2Sampler();
		}
		else if (samplerType == "regular3x3")
		{
			sampler = new Regular3x3Sampler();
		}
		else if (samplerType == "regular")
		{
			int n;

			error_in_line |= readInteger(n);

			if ( ! error_in_line)
			{
				sampler = new RegularNxNSampler();
				RecursiveRayTracer::s_NumSamplesPerPixel = n;
			}
		}
		else if (samplerType == "random")
		{
			int n;

			error_in_line |= readInteger(n);

			if ( ! error_in_line)
			{
				sampler = new RandomSampler();
				RecursiveRayTracer::s_NumSamplesPerPixel = n;
			}
		}
		else if (samplerType == "jitter")
		{
			int n;

			error_in_line |= readInteger(n);

			if ( ! error_in_line)
			{
				sampler = new JitterSampler();
				RecursiveRayTracer::s_NumSamplesPerPixel = n;
			}
		}
		else if (samplerType == "multi_jitter")
		{
			int n;

			error_in_line |= readInteger(n);

			if ( ! error_in_line)
			{
				sampler = new MultiJitterSampler();
				RecursiveRayTracer::s_NumSamplesPerPixel = n;
			}
		}
		else
		{
			reportError( "Unsupported pixel sampler type" );
			error_in_line = true;
		}

		if ( ! error_in_line)
			pixel_sampler = sampler;
	}
	else if ( ! strcmp("pixel_filter", cmd))
	{
		string filterType;

		error_in_line |= readString(filterType);

		Filter *filter = 0;

		if (filterType == "box")
		{
			filter = new BoxFilter();
		}
		else
		{
			reportError( "Unsupported pixel filter type" );
			error_in_line = true;
		}

		if ( ! error_in_line)
			pixel_filter = filter;
	}
	else if ( ! strcmp("sphere", cmd))
	{
		vec3 center;
		float radius;

		error_in_line |= readVector(center);
		error_in_line |= readFloat(radius);
		error_in_line |= requireEol();

		if ( ! error_in_line)
		{
			checkCurrentMaterial();

			objects.push_back( Sphere::CreateGeometricObject(center, radius, current_material) );

			//current_material = 0;
		}
	}
	else if ( ! strcmp("sphere_tex", cmd))
	{
		vec3 center;
		float radius;

		error_in_line |= readVector(center);
		error_in_line |= readFloat(radius);
		error_in_line |= requireEol();

		if ( ! error_in_line)
		{
			checkCurrentMaterial();

			objects.push_back( TexturedSphere::CreateGeometricObject(center, radius, current_material) );

			//current_material = 0;
		}
	}
	else if ( ! strcmp("sphere_solid_tex", cmd))
	{
		vec3 center;
		float radius;

		error_in_line |= readVector(center);
		error_in_line |= readFloat(radius);
		error_in_line |= requireEol();

		if ( ! error_in_line)
		{
			checkCurrentMaterial();

			objects.push_back( SolidTexturedSphere::CreateGeometricObject(center, radius, current_material) );

			//current_material = 0;
		}
	}
	else if ( ! strcmp("plane", cmd))
	{
		vec3 normal;
		float distance;

		error_in_line |= readVector(normal);
		error_in_line |= readFloat(distance);
		error_in_line |= requireEol();

		if ( ! error_in_line)
		{
			checkCurrentMaterial();

			objects.push_back( Plane::CreateGeometricObject(normal, distance, current_material) );

			//current_material = 0;
		}
	}
	else if ( ! strcmp("plane_tex", cmd))
	{
		vec3 normal;
		float scale, distance;

		error_in_line |= readFloat(scale);
		error_in_line |= readVector(normal);
		error_in_line |= readFloat(distance);
		error_in_line |= requireEol();

		if ( ! error_in_line)
		{
			checkCurrentMaterial();

			TexturedPlane *p = TexturedPlane::CreateGeometricObject(normal, distance, current_material);
			p->setTextureScale(scale);

			objects.push_back( p );

			//current_material = 0;
		}
	}
	else if ( ! strcmp("triangle", cmd))
	{
		vec3 v0, v1, v2;

		error_in_line |= readVector(v0);
		error_in_line |= readVector(v1);
		error_in_line |= readVector(v2);
		error_in_line |= requireEol();

		if ( ! error_in_line)
		{
			checkCurrentMaterial();

			objects.push_back( Triangle::CreateGeometricObject(v0, v1, v2, current_material) );

			//current_material = 0;
		}
	}
	else if ( ! strcmp("ambient", cmd))
	{
		vec3 ambient;

		error_in_line |= readVector(ambient);
		error_in_line |= requireEol();

		if ( ! error_in_line)
		{
			//if (current_material) delete current_material;
			//current_material = new Material(ambient);
			current_material = Material::CreateMaterial(ambient);
		}
	}
	else if ( ! strcmp("diffuse", cmd))
	{
		vec3 diffuse, ambient;

		error_in_line |= readVector(diffuse);
		error_in_line |= readVector(ambient);
		error_in_line |= requireEol();

		if ( ! error_in_line)
		{
			//DiffuseMaterial *m = new DiffuseMaterial();
			DiffuseMaterial *m = DiffuseMaterial::CreateMaterial();
			m->setDiffuseCoeff( diffuse );
			m->setAmbientCoeff( ambient );

			//if (current_material) delete current_material;
			current_material = m;
		}
	}
	else if ( ! strcmp("phong", cmd))
	{
		vec3 specular, diffuse, ambient;
		float shininess;

		error_in_line |= readVector(specular);
		error_in_line |= readFloat(shininess);
		error_in_line |= readVector(diffuse);
		error_in_line |= readVector(ambient);
		error_in_line |= requireEol();

		if ( ! error_in_line)
		{
			//PhongMaterial *m = new PhongMaterial();
			PhongMaterial *m = PhongMaterial::CreateMaterial();
			m->setPhongCoeff( specular );
			m->setShininess( shininess );
			m->setDiffuseCoeff( diffuse );
			m->setAmbientCoeff( ambient );

			//if (current_material) delete current_material;
			current_material = m;
		}
	}
	else if ( ! strcmp("mirror", cmd))
	{
		vec3 specular;

		error_in_line |= readVector(specular);
		error_in_line |= requireEol();

		if ( ! error_in_line)
		{
			//PerfectSpecularMaterial *m = new PerfectSpecularMaterial();
			PerfectSpecularMaterial *m = PerfectSpecularMaterial::CreateMaterial();
			m->setSpecularCoeff( specular );

			//if (current_material) delete current_material;
			current_material = m;
		}
	}
	else if ( ! strcmp("refract", cmd))
	{
		float eta;
		vec3 specular;

		error_in_line |= readFloat(eta);
		error_in_line |= readVector(specular);
		error_in_line |= requireEol();

		if ( ! error_in_line)
		{
			//SpecularRefractionMaterial *m = new SpecularRefractionMaterial();
			SpecularRefractionMaterial *m = SpecularRefractionMaterial::CreateMaterial();
			m->setRefractionIndex( eta );
			m->setSpecularCoeff( specular );

			//if (current_material) delete current_material;
			current_material = m;
		}
	}
	else if ( ! strcmp("texture", cmd))
	{
		string textureType;

		error_in_line |= readString(textureType);

		Texture *tex = 0;

		if (textureType == "checker")
		{
			int Nx, Ny;
			vec3 c0, c1;

			error_in_line |= readInteger(Nx);
			error_in_line |= readInteger(Ny);
			error_in_line |= readVector(c0);
			error_in_line |= readVector(c1);

			if ( ! error_in_line)
			{
				//CheckeredTexture *t = new CheckeredTexture();
				CheckeredTexture *t = CheckeredTexture::CreateTexture();
	
				t->setUNumRepeats(Nx);
				t->setVNumRepeats(Ny);
				t->setColor0(c0);
				t->setColor1(c1);

				tex = t;
			}
		}
		else if (textureType == "solid_checker")
		{
			int Nx, Ny, Nz;
			vec3 c0, c1;

			error_in_line |= readInteger(Nx);
			error_in_line |= readInteger(Ny);
			error_in_line |= readInteger(Nz);
			error_in_line |= readVector(c0);
			error_in_line |= readVector(c1);

			if ( ! error_in_line)
			{
				//SolidCheckeredTexture *t = new SolidCheckeredTexture();
				SolidCheckeredTexture *t = SolidCheckeredTexture::CreateTexture();

				t->setUNumRepeats(Nx);
				t->setVNumRepeats(Ny);
				t->setVNumRepeats(Nz);
				t->setColor0(c0);
				t->setColor1(c1);

				tex = t;
			}
		}
		else if (textureType == "noise")
		{
			float scale;
			vec3 c0, c1;

			error_in_line |= readFloat(scale);
			error_in_line |= readVector(c0);
			error_in_line |= readVector(c1);

			if ( ! error_in_line)
			{
				//NoiseTexture *t = new NoiseTexture();
				NoiseTexture *t = NoiseTexture::CreateTexture();
	
				t->setScale(scale);
				t->setColor0(c0);
				t->setColor1(c1);

				tex = t;
			}
		}
		else if (textureType == "marble")
		{
			float scale, stripe;
			vec3 c0, c1, c2;

			error_in_line |= readFloat(scale);
			error_in_line |= readFloat(stripe);
			error_in_line |= readVector(c0);
			error_in_line |= readVector(c1);
			error_in_line |= readVector(c2);

			if ( ! error_in_line)
			{
				//MarbleTexture *t = new MarbleTexture();
				MarbleTexture *t = MarbleTexture::CreateTexture();

				t->setScale(scale);
				t->setStripesPerUnit(stripe);
				t->setColor0(c0);
				t->setColor1(c1);
				t->setColor2(c2);

				tex = t;
			}
		}
		else if (textureType == "wooden")
		{
			float scale, stripe;
			vec3 c0, c1;

			error_in_line |= readFloat(scale);
			error_in_line |= readFloat(stripe);
			error_in_line |= readVector(c0);
			error_in_line |= readVector(c1);

			if ( ! error_in_line)
			{
				//WoodenTexture *t = new WoodenTexture();
				WoodenTexture *t = WoodenTexture::CreateTexture();

				t->setScale(scale);
				t->setStripesPerUnit(stripe);
				t->setColor0(c0);
				t->setColor1(c1);

				tex = t;
			}
		}
		else if (textureType == "turbulence")
		{
			float scale;
			vec3 c0, c1;

			error_in_line |= readFloat(scale);
			error_in_line |= readVector(c0);
			error_in_line |= readVector(c1);

			if ( ! error_in_line)
			{
				//TurbulenceTexture *t = new TurbulenceTexture();
				TurbulenceTexture *t = TurbulenceTexture::CreateTexture();

				t->setScale(scale);
				t->setColor0(c0);
				t->setColor1(c1);

				tex = t;
			}
		}
		else if (textureType == "dturbulence")
		{
			float scale, deriv;
			vec3 c0, c1;

			error_in_line |= readFloat(scale);
			error_in_line |= readFloat(deriv);
			error_in_line |= readVector(c0);
			error_in_line |= readVector(c1);

			if ( ! error_in_line)
			{
				//DerivTurbulenceTexture *t = new DerivTurbulenceTexture();
				DerivTurbulenceTexture *t = DerivTurbulenceTexture::CreateTexture();

				t->setScale(scale);
				t->setDeriv(deriv);
				t->setColor0(c0);
				t->setColor1(c1);

				tex = t;
			}
		}
		else if (textureType == "image")
		{
			string filename;

			error_in_line |= readString(filename);

			if ( ! error_in_line)
			{
				//ImageTexture *t = new ImageTexture();
				ImageTexture *t = ImageTexture::CreateTexture();

				t->loadImage(filename);

				tex = t;
			}
		}
		else
		{
			reportError( "Unsupported texture type" );
			error_in_line = true;
		}

		if ( ! error_in_line)
		{
			//TexturedMaterial *m = new TexturedMaterial();
			TexturedMaterial *m = TexturedMaterial::CreateMaterial();

			if ( ! tex) cerr << __FUNCTION__ << ": texture not set" << endl;

			m->setTexture( tex );

			if ( ! checkEol() )
			{
				vec3 phong, ambient;
				float shininess;

				if ( ! readVector(phong)) m->setPhongCoeff(phong);
				if ( ! readFloat(shininess)) m->setShininess(shininess);
				if ( ! readVector(ambient)) m->setAmbientCoeff(ambient);
			}

			//if (current_material) delete current_material;
			current_material = m;
		}
		//else
		//{
		//	if (tex) delete tex;
		//}
	}
	else if ( ! strcmp("direct_light_scaling", cmd))
	{
		float scaling;

		error_in_line |= readFloat(scaling);

		if ( ! error_in_line)
			direct_light_scaling = scaling;
	}
	else if ( ! strcmp("point_light", cmd))
	{
		vec3 pos, color;

		error_in_line |= readVector(pos);
		error_in_line |= readVector(color);
		error_in_line |= requireEol();

		if ( ! error_in_line)
		{
			lights.push_back( new PointLight(pos, color) );
		}
	}
	else if ( ! strcmp("direct_light", cmd))
	{
		vec3 dir, color;

		error_in_line |= readVector(dir);
		error_in_line |= readVector(color);
		error_in_line |= requireEol();

		if ( ! error_in_line)
		{
			dir.normalize();
			color *= direct_light_scaling;
			lights.push_back( new DirectionalLight(dir, color) );
		}
	}
	else if ( ! strcmp("environment_map", cmd))
	{
		string filename;

		error_in_line |= readString(filename);
		error_in_line |= requireEol();

		if ( ! error_in_line)
		{
			if ( ! specified_envmap) specified_envmap = new EnvironmentMap();
			specified_envmap->load( filename.c_str() );
		}
	}
	else if ( ! strcmp("hdr_gamma", cmd))
	{
		float gamma;

		error_in_line |= readFloat(gamma);
		error_in_line |= requireEol();

		if ( ! error_in_line)
		{
			if ( ! specified_envmap) specified_envmap = new EnvironmentMap();
			specified_envmap->setHDRGamma( gamma );
		}
	}
	else
	{
		char msg[1024];
		sprintf(msg, "unknown command: %s", cmd);
		reportError( msg );
		return true;
	}

	return error_in_line;
}

bool SceneLoader::requireChar(char c)
{
	if (error_in_line) return true;

	char r;
	iss >> ws >> r;

	if (r != c)
	{
		char msg[1024];
		sprintf(msg, "'%c' extected, but '%c' read", c, r);
		reportError( msg );
		return true;
	}

	return false;
}

bool SceneLoader::checkEol()
{
	iss >> ws;
	return iss.eof();
}

bool SceneLoader::requireEol()
{
	if (error_in_line) return true;

	iss >> ws;

	if ( ! iss.eof())
	{
		reportError("unexpected character(s) at end of line");
		return true;
	}

	return false;
}

bool SceneLoader::readFloat(float &f)
{
	if (error_in_line) return true;

	iss >> ws >> f;

	if (iss.fail())
	{
		reportError("cannot read float value");
		return true;
	}

	return false;
}

bool SceneLoader::readInteger(int &i)
{
	if (error_in_line) return true;

	iss >> ws >> i;

	if (iss.fail())
	{
		reportError("cannot read integer value");
		return true;
	}

	return false;
}

bool SceneLoader::readVector(vec3 &v)
{
	if (error_in_line) return true;

	error_in_line |= requireChar('(');
	error_in_line |= readFloat(v.x);
	error_in_line |= requireChar(',');
	error_in_line |= readFloat(v.y);
	error_in_line |= requireChar(',');
	error_in_line |= readFloat(v.z);
	error_in_line |= requireChar(')');

	return error_in_line;
}

bool SceneLoader::readString(string &str)
{
	if (error_in_line) return true;

	iss >> ws >> str;

	if (iss.fail())
	{
		reportError("cannot read string");
		return true;
	}

	return false;
}

void SceneLoader::checkCurrentMaterial()
{
	// check current material
	if ( ! current_material)
	{
		reportWarning("no material specified, default material used");
		//DiffuseMaterial *diffuse = new DiffuseMaterial();
		DiffuseMaterial *diffuse = DiffuseMaterial::CreateMaterial();
		diffuse->setDiffuseCoeff(0.8f,0.8f,0.8f);
		diffuse->setAmbientCoeff(0.2f,0.2f,0.2f);
		current_material = diffuse;
	}
}

void SceneLoader::reportWarning(const char *str)
{
	cerr << "Warning at line " << line_number << ": " << str << endl;
}

void SceneLoader::reportError(const char *str)
{
	cerr << "Error at line " << line_number << ": " << str << endl;
}
