#include "UniformGrid.h"
#include <cstdlib>
#include <GL/glut.h>

using namespace std;
using namespace MyAlgebra;

extern bool g_SingleRayDebug;
extern vector<int> g_TracedGridCells;

bool UniformGrid::init( const MyAlgebra::vec3 &bboxMin, const MyAlgebra::vec3 &bboxMax, int nXCells, int nYCells, int nZCells )
{
	if ( nXCells <= 0 || nYCells <= 0 || nZCells <= 0 )
	{
		cerr << __FUNCTION__ << ": invalid cell size: " << nXCells << "x" << nYCells << "x" << nZCells << endl;
		return false;
	}

	if ( bboxMin.x >= bboxMax.x || bboxMin.y >= bboxMax.y || bboxMin.z >= bboxMax.z )
	{
		cerr << __FUNCTION__ << ": bboxMin coordinate(s) are larger than bboxMax coordinates: bboxMin=" << bboxMin << ", bboxMax=" << bboxMax << endl;
		return false;
	}

	m_BboxMin = bboxMin;
	m_BboxMax = bboxMax;

	m_NumXCells = nXCells;
	m_NumYCells = nYCells;
	m_NumZCells = nZCells;

	const vec3 bboxSize = bboxMax - bboxMin;

	m_DeltaX = bboxSize.x/(float)m_NumXCells;
	m_DeltaY = bboxSize.y/(float)m_NumYCells;
	m_DeltaZ = bboxSize.z/(float)m_NumZCells;

	if (m_pCells) delete [] m_pCells;

	m_pCells = new vector<GeometricObject *>[ m_NumXCells * m_NumYCells * m_NumZCells ];

	cerr << __FUNCTION__ << ": info: " << endl
		 << "    allocated cells: " << m_NumXCells << "x" << m_NumYCells << "x" << m_NumZCells << endl
		 << "    bbox: " << m_BboxMin << ", " << m_BboxMax << endl;

	return true;
}

inline int clampInt(int x, int xMin, int xMax)
{
	return (x < xMin) ? xMin : (x > xMax) ? xMax : x;
}

void UniformGrid::addObject( const GeometricObject *o )
{
	// Obh̒[猩Ăǂ
	const vec3 bboxMin = o->getBoundingBoxMin() - m_BboxMin;
	const vec3 bboxMax = o->getBoundingBoxMax() - m_BboxMin;

	//cerr << __FUNCTION__ << ": bbox: " << bboxMin << ", " << bboxMax << endl;
	//cerr << __FUNCTION__ << ": cell sizes: [" << m_DeltaX << "," << m_DeltaY << "," << m_DeltaZ << "]" << endl;

	//const int xStart = clampInt( (int)floorf(bboxMin.x/m_DeltaX), 0, m_NumXCells-1 );
	//const int xEnd = clampInt( (int)ceilf(bboxMax.x/m_DeltaX), 0,  m_NumXCells-1 );
	//const int yStart = clampInt( (int)floorf(bboxMin.y/m_DeltaY), 0, m_NumYCells-1 );
	//const int yEnd = clampInt( (int)ceilf(bboxMax.y/m_DeltaY), 0, m_NumYCells-1 );
	//const int zStart = clampInt( (int)floorf(bboxMin.z/m_DeltaZ), 0, m_NumZCells-1 );
	//const int zEnd = clampInt( (int)ceilf(bboxMax.z/m_DeltaZ), 0, m_NumZCells-1 );

	const int xStart = clampInt( (int)floorf(bboxMin.x/m_DeltaX), 0, m_NumXCells-1 );
	const int xEnd = clampInt( (int)floorf(bboxMax.x/m_DeltaX), 0,  m_NumXCells-1 );
	const int yStart = clampInt( (int)floorf(bboxMin.y/m_DeltaY), 0, m_NumYCells-1 );
	const int yEnd = clampInt( (int)floorf(bboxMax.y/m_DeltaY), 0, m_NumYCells-1 );
	const int zStart = clampInt( (int)floorf(bboxMin.z/m_DeltaZ), 0, m_NumZCells-1 );
	const int zEnd = clampInt( (int)floorf(bboxMax.z/m_DeltaZ), 0, m_NumZCells-1 );

	//cerr << __FUNCTION__ << ": range: [" << xStart << "," << xEnd << "], [" << yStart << "," << yEnd
	//	 << "], [" << zStart << "," << zEnd << "]" << endl;

	for (int zi=zStart; zi<=zEnd; zi++)
	{
		for (int yi=yStart; yi<=yEnd; yi++)
		{
			for (int xi=xStart; xi<=xEnd; xi++)
			{
				getCell(xi,yi,zi).push_back( const_cast<GeometricObject *>(o) );
			}
		}
	}

	//const int nCells = (xEnd-xStart+1)*(yEnd-yStart+1)*(zEnd-zStart+1);
	//cerr << __FUNCTION__ << ": registered in " << nCells << " cells" << endl;
}

void UniformGrid::drawWireframe() const
{
	if ( ! m_pCells) return;

	glLineWidth(1.f);

	const float xEnd = m_BboxMin.x + m_DeltaX * m_NumXCells;
	const float yEnd = m_BboxMin.y + m_DeltaY * m_NumYCells;
	const float zEnd = m_BboxMin.z + m_DeltaZ * m_NumZCells;

	// z ̐

	glBegin(GL_LINES);
	for (int yi=0; yi<=m_NumYCells; yi++)
	{
		const float y = m_BboxMin.y + m_DeltaY * yi;

		for (int xi=0; xi<=m_NumXCells; xi++)
		{
			const float x = m_BboxMin.x + m_DeltaX * xi;

			glVertex3f(x, y, m_BboxMin.z);
			glVertex3f(x, y, zEnd);
		}
	}
	glEnd();

	// y ̐

	glBegin(GL_LINES);
	for (int zi=0; zi<=m_NumZCells; zi++)
	{
		const float z = m_BboxMin.z + m_DeltaZ * zi;

		for (int xi=0; xi<=m_NumXCells; xi++)
		{
			const float x = m_BboxMin.x + m_DeltaX * xi;

			glVertex3f(x, m_BboxMin.y, z);
			glVertex3f(x, yEnd, z);
		}
	}
	glEnd();

	// x ̐

	glBegin(GL_LINES);
	for (int zi=0; zi<=m_NumZCells; zi++)
	{
		const float z = m_BboxMin.z + m_DeltaZ * zi;

		for (int yi=0; yi<=m_NumYCells; yi++)
		{
			const float y = m_BboxMin.y + m_DeltaY * yi;

			glVertex3f(m_BboxMin.x, y, z);
			glVertex3f(xEnd, y, z);
		}
	}
	glEnd();


	if ( ! g_TracedGridCells.empty() )
	{
		glLineWidth( 5.f );
		glColor3f(1,0,0);

		for (int i=0; i<(int)g_TracedGridCells.size(); i++)
		{
			drawCellWireframe( g_TracedGridCells[i] );
		}
	}
}

void UniformGrid::drawCellWireframe(int xi, int yi, int zi) const
{
	if ( ! m_pCells) return;

	const float x0 = m_BboxMin.x + m_DeltaX * xi;
	const float y0 = m_BboxMin.y + m_DeltaY * yi;
	const float z0 = m_BboxMin.z + m_DeltaZ * zi;

	const float x1 = x0 + m_DeltaX;
	const float y1 = y0 + m_DeltaY;
	const float z1 = z0 + m_DeltaZ;

	glBegin(GL_LINE_LOOP);
	glVertex3f(x0, y0, z0);
	glVertex3f(x1, y0, z0);
	glVertex3f(x1, y1, z0);
	glVertex3f(x0, y1, z0);
	glEnd();

	glBegin(GL_LINE_LOOP);
	glVertex3f(x0, y0, z1);
	glVertex3f(x1, y0, z1);
	glVertex3f(x1, y1, z1);
	glVertex3f(x0, y1, z1);
	glEnd();

	glBegin(GL_LINES);
	glVertex3f(x0, y0, z0);	glVertex3f(x0, y0, z1);
	glVertex3f(x1, y0, z0);	glVertex3f(x1, y0, z1);
	glVertex3f(x1, y1, z0);	glVertex3f(x1, y1, z1);
	glVertex3f(x0, y1, z0);	glVertex3f(x0, y1, z1);
	glEnd();
}

void UniformGrid::drawCellWireframe(int idx) const
{
	if ( ! m_pCells) return;

	const int xi = idx % m_NumXCells;
	const int yi = (idx - xi) / m_NumXCells % m_NumYCells;
	const int zi = ((idx - xi) / m_NumXCells - yi) / m_NumYCells;

	drawCellWireframe(xi, yi, zi);
}

