#pragma once

#include <iostream>

namespace UbyteLib {

typedef unsigned char ubyte;
typedef float Real;

struct ubyte2;
struct ubyte3;
struct ubyte4;

struct ubyte2
{
    union {
		struct {
			ubyte x,y;        // standard names for components
		};
		struct {
			ubyte s,t;
		};
        ubyte vec_array[2];     // array access
    };

	ubyte2() { }
	ubyte2(const ubyte v) : x(v), y(v) {}
    ubyte2(const ubyte _x, const ubyte _y) : x(_x), y(_y) { }
    ubyte2(const ubyte* xy) : x(xy[0]), y(xy[1]) { }
	ubyte2(const ubyte2& u) : x(u.x), y(u.y) { }
	ubyte2(const ubyte2& u1, const ubyte2& u2)
		: x(u2.x - u1.x), y(u2.y - u1.y) {}

	inline void set(const ubyte val)
	{
		x = val;
		y = val;
	}

	inline void set(const ubyte _x, const ubyte _y)
	{
		x = _x;
		y = _y;
	}

	inline void set(const ubyte2 &v)
	{
		x = v.x;
		y = v.y;
	}

	inline void add(const ubyte2 &u)
	{
		x += u.x;
		y += u.y;
	}

	inline void sub(const ubyte2 &u)
	{
		x -= u.x;
		y -= u.y;
	}

	inline void scale(const Real val)
	{
		x = (ubyte)(x*val);
		y = (ubyte)(y*val);
	}

	inline void lerp(const Real t, const ubyte2 & u, const ubyte2 & v)
	{
		x = (ubyte)(u.x + t * (v.x - u.x));
		y = (ubyte)(u.y + t * (v.y - u.y));
	}

	inline ubyte2& operator=(const ubyte2 & u)
	{
		x = u.x;
		y = u.y;
        return *this;
	}

    inline bool operator==(const ubyte2 & u) const
    {
        return (u.x == x && u.y == y);
    }

    inline bool operator!=( const ubyte2& rhs ) const
    {
        return !(*this == rhs );
    }

    inline ubyte2 & operator*=(const Real lambda)
    {
        x = ubyte(x*lambda);
		y = ubyte(y*lambda);
        return *this;
    }

	inline ubyte2 & operator/=(const Real lambda)
    {
        x = ubyte(x/lambda);
        y = ubyte(y/lambda);
        return *this;
    }

    inline ubyte2 operator - () const
	{
		return ubyte2(-x, -y);
	}

	inline ubyte2 & operator-=(const ubyte2 & u)
    {
        x -= u.x;
        y -= u.y;
        return *this;
    }

    inline ubyte2 & operator+=(const ubyte2 & u)
    {
        x += u.x;
        y += u.y;
        return *this;
    }

    inline ubyte & operator[](const int i)
    {
        return vec_array[i];
    }

    inline const ubyte operator[](const int i) const
    {
        return vec_array[i];
    }

	inline operator const ubyte *() const { return vec_array; }

	inline friend ubyte dot(const ubyte2 & u, const ubyte2 & v)
	{
		return u.x*v.x + u.y*v.y;
	}

	inline friend ubyte2 & madd(ubyte2 & u, const ubyte2& v, Real lambda)
	{
	    u.x += ubyte(v.x*lambda);
	    u.y += ubyte(v.y*lambda);
		return u;
	}

	inline friend ubyte2 & mult(ubyte2 & u, const ubyte2& v, Real lambda)
	{
	    u.x = ubyte(v.x*lambda);
	    u.y = ubyte(v.y*lambda);
		return u;
	}

	inline friend ubyte2 & mult(ubyte2 & u, Real lambda, const ubyte2& v)
	{
	    u.x = ubyte(v.x*lambda);
	    u.y = ubyte(v.y*lambda);
		return u;
	}

	inline friend ubyte2 & mult(ubyte2 & u, const ubyte2& v, const ubyte2& w)
	{
	    u.x = v.x*w.x;
	    u.y = v.y*w.y;
		return u;
	}

	inline friend ubyte2 & sub(ubyte2 & u, const ubyte2& v, const ubyte2& w)
	{
	    u.x = v.x - w.x;
	    u.y = v.y - w.y;
		return u;
	}

	inline friend ubyte2 & add(ubyte2 & u, const ubyte2& v, const ubyte2& w)
	{
	    u.x = v.x + w.x;
	    u.y = v.y + w.y;
		return u;
	}

	inline friend ubyte2 & scale(ubyte2& u, Real s)
	{
	    u.x = ubyte(u.x*s);
	    u.y = ubyte(u.y*s);
		return u;
	}

	inline friend ubyte2 & scale(Real s, ubyte2& u)
	{
	    u.x = ubyte(u.x*s);
	    u.y = ubyte(u.y*s);
		return u;
	}

	inline friend ubyte2 & lerp(ubyte2 & w, Real t, const ubyte2 & u, const ubyte2 & v)
	{
		w.lerp(t, u, v);
		return w;
	}

	inline friend const ubyte2 operator+(const ubyte2& u, const ubyte2& v)
	{
		return ubyte2(u.x + v.x, u.y + v.y);
	}

	inline friend const ubyte2 operator-(const ubyte2& u, const ubyte2& v)
	{
	    return ubyte2(u.x - v.x, u.y - v.y);
	}

	inline friend const ubyte2 operator*(Real s, const ubyte2& u)
	{
		return ubyte2(ubyte(s * u.x), ubyte(s * u.y));
	}

	inline friend const ubyte2 operator*(const ubyte2& u, Real s)
	{
		return ubyte2(ubyte(s * u.x), ubyte(s * u.y));
	}

	inline friend const ubyte2 operator/(const ubyte2& u, Real s)
	{
		return ubyte2((ubyte)(u.x / s), (ubyte)(u.y / s));
	}

	inline friend ubyte operator*(const ubyte2& u, const ubyte2& v)
	{
		return u.x * v.x + u.y * v.y;
	}

	inline friend std::ostream& operator<<(std::ostream &o, const ubyte2& u)
	{
		return o << "(" << (int)u.x << "," << (int)u.y << ")";
	}

};

struct ubyte3
{
    union {
		struct {
			ubyte x,y,z;        // standard names for components
		};
		struct {
			ubyte r,g,b;
		};
        ubyte vec_array[3];     // array access
    };

	ubyte3() { }
	ubyte3(const ubyte v) : x(v), y(v), z(v) {}
    ubyte3(const ubyte _x, const ubyte _y, const ubyte _z)
		: x(_x), y(_y), z(_z) { }
    ubyte3(const ubyte* xyz) : x(xyz[0]), y(xyz[1]), z(xyz[2]) { }
	ubyte3(const ubyte3& u) : x(u.x), y(u.y), z(u.z) { }
	ubyte3(const ubyte3& u1, const ubyte3& u2)
		: x(u2.x - u1.x), y(u2.y - u1.y), z(u2.z - u1.z) {}

	inline void set(const ubyte val)
	{
		x = val;
		y = val;
		z = val;
	}

	inline void set(const ubyte _x, const ubyte _y, const ubyte _z)
	{
		x = _x;
		y = _y;
		z = _z;
	}

	inline void set(const ubyte3 &v)
	{
		x = v.x;
		y = v.y;
		z = v.z;
	}

	inline void add(const ubyte3 &u)
	{
		x += u.x;
		y += u.y;
		z += u.z;
	}

	inline void sub(const ubyte3 &u)
	{
		x -= u.x;
		y -= u.y;
		z -= u.z;
	}

	inline void scale(const Real val)
	{
		x = ubyte(x*val);
		y = ubyte(y*val);
		z = ubyte(z*val);
	}

	inline void lerp(const Real t, const ubyte3 & u, const ubyte3 & v)
	{
		x = ubyte(u.x + t * (v.x - u.x));
		y = ubyte(u.y + t * (v.y - u.y));
		z = ubyte(u.z + t * (v.z - u.z));
	}

	inline ubyte3& operator=(const ubyte3 & u)
	{
		x = u.x; y = u.y; z = u.z;
        return *this;
	}

    inline bool operator==(const ubyte3 & u) const
    {
        return (u.x == x && u.y == y && u.z == z);
    }

    inline bool operator!=( const ubyte3& rhs ) const
    {
        return !(*this == rhs );
    }

    inline ubyte3 & operator*=(const Real lambda)
    {
        x = ubyte(x*lambda);
        y = ubyte(y*lambda);
        z = ubyte(z*lambda);
        return *this;
    }

	inline ubyte3 & operator/=(const Real lambda)
    {
		const Real invLambda = 1/lambda;
        x = ubyte(x*invLambda);
        y = ubyte(y*invLambda);
        z = ubyte(z*invLambda);
        return *this;
    }

	inline ubyte3 operator - () const
	{
		return ubyte3(-x, -y, -z);
	}

    inline ubyte3 & operator-=(const ubyte3 & u)
    {
        x -= u.x;
        y -= u.y;
        z -= u.z;
        return *this;
    }

    inline ubyte3 & operator+=(const ubyte3 & u)
    {
        x += u.x;
        y += u.y;
        z += u.z;
        return *this;
    }

    inline ubyte & operator[](int i)
    {
        return vec_array[i];
    }

    inline const ubyte operator[](int i) const
    {
        return vec_array[i];
    }

	inline operator const ubyte *() const { return vec_array; }

	inline friend ubyte dot(const ubyte3 & u, const ubyte3 & v)
	{
		return u.x*v.x + u.y*v.y + u.z*v.z;
	}

	inline friend ubyte3 & madd(ubyte3 & u, const ubyte3& v, const Real lambda)
	{
	    u.x += ubyte(v.x*lambda);
	    u.y += ubyte(v.y*lambda);
	    u.z += ubyte(v.z*lambda);
		return u;
	}

	inline friend ubyte3 & mult(ubyte3 & u, const ubyte3& v, const Real lambda)
	{
	    u.x = ubyte(v.x*lambda);
	    u.y = ubyte(v.y*lambda);
	    u.z = ubyte(v.z*lambda);
		return u;
	}

	inline friend ubyte3 & mult(ubyte3 & u, const Real lambda, const ubyte3& v)
	{
	    u.x = ubyte(v.x*lambda);
	    u.y = ubyte(v.y*lambda);
	    u.z = ubyte(v.z*lambda);
		return u;
	}

	inline friend ubyte3 & mult(ubyte3 & u, const ubyte3& v, const ubyte3& w)
	{
	    u.x = v.x*w.x;
	    u.y = v.y*w.y;
	    u.z = v.z*w.z;
		return u;
	}

	inline friend ubyte3 & sub(ubyte3 & u, const ubyte3& v, const ubyte3& w)
	{
	    u.x = v.x - w.x;
	    u.y = v.y - w.y;
	    u.z = v.z - w.z;
		return u;
	}

	inline friend ubyte3 & add(ubyte3 & u, const ubyte3& v, const ubyte3& w)
	{
	    u.x = v.x + w.x;
	    u.y = v.y + w.y;
	    u.z = v.z + w.z;
		return u;
	}

	inline friend ubyte3 & scale(ubyte3& u, const Real s)
	{
		u.x = ubyte(u.x*s);
		u.y = ubyte(u.y*s);
		u.z = ubyte(u.z*s);
		return u;
	}

	inline friend ubyte3 & scale(const Real s, ubyte3& u)
	{
		u.x = ubyte(u.x*s);
		u.y = ubyte(u.y*s);
		u.z = ubyte(u.z*s);
		return u;
	}

	inline friend ubyte3 & lerp(ubyte3 & w, Real t, const ubyte3 & u, const ubyte3 & v)
	{
		w.lerp(t, u, v);
		return w;
	}

	inline friend const ubyte3 operator+(const ubyte3& u, const ubyte3& v)
	{
		return ubyte3(u.x + v.x, u.y + v.y, u.z + v.z);
	}

	inline friend const ubyte3 operator-(const ubyte3& u, const ubyte3& v)
	{
	    return ubyte3(u.x - v.x, u.y - v.y, u.z - v.z);
	}

	inline friend const ubyte3 operator^(const ubyte3& u, const ubyte3& v)
	{
	    return ubyte3(u.y*v.z - u.z*v.y, u.z*v.x - u.x*v.z, u.x*v.y - u.y*v.x);
	}

	inline friend const ubyte3 operator*(const Real s, const ubyte3& u)
	{
		return ubyte3(ubyte(s*u.x), ubyte(s*u.y), ubyte(s*u.z));
	}

	inline friend const ubyte3 operator*(const ubyte3& u, const Real s)
	{
		return ubyte3(ubyte(s*u.x), ubyte(s*u.y), ubyte(s*u.z));
	}

	inline friend const ubyte3 operator/(const ubyte3& u, const Real s)
	{
		const Real invS = 1/s;
		return ubyte3(ubyte(u.x*invS), ubyte(u.y*invS), ubyte(u.z*invS));
	}

	inline friend ubyte operator*(const ubyte3& u, const ubyte3& v)
	{
		return u.x * v.x + u.y * v.y + u.z * v.z;
	}

	inline friend std::ostream& operator<<(std::ostream &o, const ubyte3& u)
	{
		return o << "(" << (int)u.x << "," << (int)u.y << "," << (int)u.z << ")";
	}

};

struct ubyte4
{
    union {
		struct {
			ubyte x,y,z,w;        // standard names for components
		};
		struct {
			ubyte r,g,b,a;
		};
        ubyte vec_array[4];     // array access
    };

	ubyte4() { }
	ubyte4(const ubyte v) : x(v), y(v), z(v), w(v) {}
    ubyte4(ubyte _x, ubyte _y, ubyte _z, ubyte _w)
		: x(_x), y(_y), z(_z), w(_w) { }
    ubyte4(const ubyte* xyzw) : x(xyzw[0]), y(xyzw[1]), z(xyzw[2]), w(xyzw[3]) { }
	ubyte4(const ubyte4& u) : x(u.x), y(u.y), z(u.z), w(u.w) { }
	ubyte4(const ubyte4& _from, const ubyte4& _to)
		: x(_to.x - _from.x), y(_to.y - _from.y), z(_to.z - _from.z), w(_to.w - _from.w) {}

	inline void set(const ubyte val)
	{
		x = val;
		y = val;
		z = val;
		w = val;
	}

	inline void set(const ubyte _x, const ubyte _y, const ubyte _z, const ubyte _w)
	{
		x = _x;
		y = _y;
		z = _z;
		w = _w;
	}

	inline void set(const ubyte4 &v)
	{
		x = v.x;
		y = v.y;
		z = v.z;
		w = v.w;
	}

	inline void add(const ubyte4 &u)
	{
		x += u.x;
		y += u.y;
		z += u.z;
		w += u.w;
	}

	inline void sub(const ubyte4 &u)
	{
		x -= u.x;
		y -= u.y;
		z -= u.z;
		w -= u.w;
	}

	inline void scale(const Real val)
	{
		x = ubyte(x*val);
		y = ubyte(y*val);
		z = ubyte(z*val);
		w = ubyte(w*val);
	}

	inline void lerp(const Real t, const ubyte4 & u, const ubyte4 & v)
	{
		x = ubyte(u.x + t * (v.x - u.x));
		y = ubyte(u.y + t * (v.y - u.y));
		z = ubyte(u.z + t * (v.z - u.z));
		w = ubyte(u.w + t * (v.w - u.w));
	}

	inline ubyte4& operator=(const ubyte4 & u)
	{
		x = u.x; y = u.y; z = u.z; w = u.w;
        return *this;
	}

    inline bool operator==(const ubyte4 & u) const
    {
        return (u.x == x && u.y == y && u.z == z && u.w == w);
    }

    inline bool operator!=( const ubyte4& rhs ) const
    {
        return !(*this == rhs );
    }

    inline ubyte4 & operator*=(const Real lambda)
    {
		x = ubyte(x*lambda);
		y = ubyte(y*lambda);
		z = ubyte(z*lambda);
		w = ubyte(w*lambda);
		return *this;
    }

	inline ubyte4 & operator/=(const Real lambda)
    {
		const Real invLambda = 1/lambda;
		x = ubyte(x*invLambda);
		y = ubyte(y*invLambda);
		z = ubyte(z*invLambda);
		w = ubyte(w*invLambda);
		return *this;
    }

	inline ubyte4 operator - () const
	{
		return ubyte4(-x, -y, -z, -w);
	}

    inline ubyte4 & operator-=(const ubyte4 & u)
    {
		x -= u.x;
		y -= u.y;
		z -= u.z;
		w -= u.w;
		return *this;
    }

    inline ubyte4 & operator+=(const ubyte4 & u)
    {
		x += u.x;
		y += u.y;
		z += u.z;
		w += u.w;
		return *this;
    }

    inline ubyte & operator[](int i)
    {
        return vec_array[i];
    }

    inline const ubyte operator[](int i) const
    {
        return vec_array[i];
    }

	inline operator const ubyte *() const { return vec_array; }

	inline friend ubyte4 & madd(ubyte4 & u, const ubyte4& v, const Real lambda)
	{
		u.x += ubyte(v.x*lambda);
		u.y += ubyte(v.y*lambda);
		u.z += ubyte(v.z*lambda);
		u.w += ubyte(v.w*lambda);
		return u;
	}

	inline friend ubyte4 & mult(ubyte4 & u, const ubyte4& v, const Real lambda)
	{
		u.x = ubyte(v.x*lambda);
		u.y = ubyte(v.y*lambda);
		u.z = ubyte(v.z*lambda);
		u.w = ubyte(v.w*lambda);
		return u;
	}

	inline friend ubyte4 & mult(ubyte4 & u, const Real lambda, const ubyte4& v)
	{
	    u.x = ubyte(v.x*lambda);
	    u.y = ubyte(v.y*lambda);
	    u.z = ubyte(v.z*lambda);
	    u.w = ubyte(v.w*lambda);
		return u;
	}

	inline friend ubyte4 & mult(ubyte4 & u, const ubyte4& v, const ubyte4& w)
	{
	    u.x = v.x*w.x;
	    u.y = v.y*w.y;
	    u.z = v.z*w.z;
	    u.w = v.w*w.w;
		return u;
	}

	inline friend ubyte4 & sub(ubyte4 & u, const ubyte4& v, const ubyte4& w)
	{
		u.x = v.x - w.x;
		u.y = v.y - w.y;
		u.z = v.z - w.z;
		u.w = v.w - w.w;
		return u;
	}

	inline friend ubyte4 & add(ubyte4 & u, const ubyte4& v, const ubyte4& w)
	{
	    u.x = v.x + w.x;
	    u.y = v.y + w.y;
	    u.z = v.z + w.z;
	    u.w = v.w + w.w;
		return u;
	}

	inline friend ubyte4 & scale(ubyte4& u, const Real s)
	{
		u.x = ubyte(u.x*s);
		u.y = ubyte(u.y*s);
		u.z = ubyte(u.z*s);
		u.w = ubyte(u.w*s);
		return u;
	}

	inline friend ubyte4 & scale(const Real s, ubyte4& u)
	{
		u.x = ubyte(u.x*s);
		u.y = ubyte(u.y*s);
		u.z = ubyte(u.z*s);
		u.w = ubyte(u.w*s);
		return u;
	}

	inline friend ubyte4 & lerp(ubyte4 & w, const Real t, const ubyte4 & u, const ubyte4 & v)
	{
		w.lerp(t, u, v);
		return w;
	}

	inline friend const ubyte4 operator+(const ubyte4& u, const ubyte4& v)
	{
		return ubyte4(u.x + v.x, u.y + v.y, u.z + v.z, u.w + v.w);
	}

	inline friend const ubyte4 operator-(const ubyte4& u, const ubyte4& v)
	{
	    return ubyte4(u.x - v.x, u.y - v.y, u.z - v.z, u.w - v.w);
	}

	inline friend const ubyte4 operator*(const Real s, const ubyte4& u)
	{
		return ubyte4(ubyte(s*u.x), ubyte(s*u.y), ubyte(s*u.z), ubyte(s*u.w));
	}

	inline friend const ubyte4 operator/(const ubyte4& u, const Real s)
	{
		const Real invS = 1/s;
		return ubyte4(ubyte(u.x*invS), ubyte(u.y*invS), ubyte(u.z*invS), ubyte(u.w*invS));
	}

	inline friend ubyte operator*(const ubyte4& u, const ubyte4& v)
	{
		return (u.x * v.x + u.y * v.y + u.z * v.z) / (u.w * v.w);
	}

	inline friend std::ostream& operator<<(std::ostream &o, const ubyte4& u)
	{
		return o << "(" << (int)u.x << "," << (int)u.y << "," << (int)u.z << "," << (int)u.w << ")";
	}

};

}

